; GPS Board Code V1.0
; SERVO Magazine version
; See "Quick Start" file for operational notes - jwm 2008-2009
;*******************************************************************************
; LCD E mod for John's board		18 Mar 09  (complete)

;********************************************************************************
		LIST P=16F876A
		INCLUDE <P16F876A.INC>
		__CONFIG _WDT_OFF & _HS_OSC & _BODEN_OFF
		
		ERRORLEVEL -302					; kill off the Banksel messages
		ERRORLEVEL -305					; kill linker general msgs
		ERRORLEVEL -306					; and boundary msgs
		ERRORLEVEL -202					; boundary msgs
; *********************  BAUD RATE  *********************************************

#define FOSC D'20000000'				; USER defines this Rock Freq.
#define BAUD D'4800'					; USER decides this!

#define DIVIDER (FOSC/(D'64' * BAUD))-1 ; Result will to to BRG


TMR0_Led	equ	2						; LED on RA2 for Heartbeat
TMR0_Rate	equ	0xFE					; base heartbeat rate

;#define		DEBUG	1				; comment to KILL DEBUG AREA
#define		LCD_Main 1					; main LCD nav output
#define		ISSUE_RS232	equ	1			; comment to kill RS-232 output

; If the moving average calcs are NOT desired, make NOMA (below) 1.  To ALLOW the MA
; calcs, make NOMA 0.

NOMA		equ	1						; 1=NO MA, 0=ALLOW MA
; ******************************************************************************
; ------------------------------------------------------------------------------
;						MACROs For Bank Switching
; ------------------------------------------------------------------------------
Bank_0		macro					; Total set up for Page 0
			BANKSEL	PORTA
			bcf		STATUS,IRP
			endm

Bank_1		macro					; Page 1
			BANKSEL	TRISA
			bcf		STATUS,IRP
			endm
Bank_2		macro					; Set up for Page 2
			BANKSEL	EEDATA
			bsf		STATUS,IRP
			endm

; -------------------------------------------------------------------------------
;                       LCD Equates
; Specific to the LCD
; ------------------------------------------------------------------------------- 
			radix	dec		; Decimal radix for initialization

LCD_DATA	EQU	PORTB		; LCD data lines interface
LCD_DATA_TRIS	EQU	TRISB
LCD_CTRL	EQU	PORTA		; LCD control lines interface

LCD_LINE1	EQU	0x000		; Start of Line1 (within controller's RAM !)
LCD_LINE2	EQU	0x040		; Start of Line2 (deep in bowls of Hitachi controller !)

LCD_CLEAR	equ	1			; clear entire display

; PORTA bits
LCD_E		EQU	3		; LCD Enable control line		~~Was 2, this for John Board
LCD_RW		EQU	1		; LCD Read/Write control line
LCD_RS		EQU	0		; LCD Register-Select control line

; PORTB bits
DB7		EQU	7		; LCD bus 7 (MSB)
DB6		EQU	6		
DB5		EQU	5		
DB4		EQU	4		
DB3		EQU	3		
DB2		EQU	2		
DB1		EQU	1		
DB0		EQU	0		; LCD bus 0 (LSB)

; Following LEDS share b0-b7 of the LCD Data Bus

Yel_Led	  EQU 0		; Yellow LED on LCD Board (GPS Lock OK)
Red_M_Led EQU 2		; Middle Red LED (GPS Sentence received)
Red_T_Led EQU 1		; Top RED LED (SPI sync with Nav Board)

; Following STATUS LEDS on the rear of GPS Board

Five_Sec_Led equ 	3		; GPS sentence five second indicator LED (NOW 3 SEC. after trials)
CS_Error	equ		4		; GPS sentence checksum error, sentence ignored
Lost_Sync	equ		5		; Lost GPS<->NAV Board sync (fatal)
GPS_Overflow equ	6		; GPS Sentence Buffer tried to overflow(not fatal)
Heart_Led equ	    7		; Heartbeat LED on PB(7)

; Following refers to the buttons S1-S4 on the LCD Board

SW4		EQU	3		; Rightmost button (S4)
SW3		EQU	2		; Right, Inner button (S3)
SW2		EQU	1		; Left, Inner button (S2)
SW1		EQU	0		; Leftmost button	(S1)
;
U_But_S2 EQU 7		; SW2 depressed flag
BL_In   EQU  7		; Within Buttons and LEDS function (lockout)
;
; ****************************************************************

;--------------------------------------------------------------------------------
; 						General Equates
; Allocate SRAM as required
; NOTE: Following to be converted to "CBLOCK" form once code finalized!
;--------------------------------------------------------------------------------

		radix	dec
		org		0x20						; Start of Bank0 SRAM
Page0_B	equ		0x20
Bank0_L equ		0x20

Cntrlo	equ		Page0_B						; Timer accumulators low byte
Cntrhi	equ		Page0_B+1					; High byte
A_Shadow equ	Page0_B+2					; PORTA output shadow byte
;
S_Temp	equ		Page0_B+3					; Save CC in ISR
P_Temp  equ		Page0_B+4					; Save PCLATH in ISR
F_Temp	equ		Page0_B+5					; Save FSR over ISR
Gen_Temp equ	Page0_B+6					; general Page 0 temp
;
; Following used by the UART TX routines
;
TX_String 	equ	Page0_B+8				; 40 locations reserved
TX_S_Ptr	equ	Page0_B+48				; ptr into above
TX_Work_Ptr	  equ	Page0_B+49			; running ptr
Temp_0		equ	Page0_B+50				; Used by "Small Delay"
Temp_1		equ	Page0_B+51				; Used by "Small Delay" union w/below
;
; Following Page0_B locations used by the LCD Module
;
LCD_TEMP	equ	Bank0_L+51		; LCD subroutines internal use
COUNT		equ	Bank0_L+52		; A counter (also Dly0)
DELAY		equ	Bank0_L+53		; Used in DELAYxxx routines (also Dly1)
X_DELAY		equ	Bank0_L+54		; Used in X_DELAYxxx routines (also DLY2)
;
RX_Char		equ	Bank0_L+55		; RX char from ISR
Line_Position_Cntr equ Bank0_L+56 ; Position in a line
Line_Ctl equ	Bank0_L+57		; Line toggle control
count	equ		Bank0_L+58
Worker	equ		Bank0_L+59		; for general use (also Temp) ~
C_Shadow equ	Bank0_L+60		; PORTC Buttons and LEDS
C_Debounce equ	Bank0_L+61		; button debounce control
Panel_Buts equ	Bank0_L+62		; Button Flags
;
; Following dedicated to SSI Transfer operations
;
ST_Wrk0		equ	Bank0_L+63
ST_Wrk1		equ	Bank0_L+64
ST_Wrk2		equ	Bank0_L+65
ST_Wrk3		equ	Bank0_L+66
MA_Ready	equ	Bank0_L+67		; have a valid MA
;
Five_Sec	equ	Bank0_L+68		; about 5 sec counter driven by GPS sentence ticks
Sim_Sec		equ	Bank0_L+69		; counter for Sim "Play" command
;
; Unions
;
Dly0	equ		COUNT
Dly1	equ		DELAY
Dly2	equ		X_DELAY			; ~ this and above will disappear ~
Temp	equ		Worker
; **********************************************************************************
;					END Of BANK 0
; **********************************************************************************
;
;					Following in Bank 1 SRAM
; Here is the main GPS Sentence input buffer and control words, flags, etc
; 80 bytes max in this block
;
			org		0xA0
Page1_B		equ		0xA0
GPS_Raw		equ		Page1_B				; 70D locations for GPS sentence
GPS_S_Ptr	equ		Page1_B+70			; ptr to above
GPS_W_Ptr	equ		Page1_B+71			; ISR working ptr
GPS_Flags	equ		Page1_B+72
GPS_Recent	equ		Page1_B+73			; most recent received char
GPS_FSR_Sav equ		Page1_B+74			; FSR temp for Page 1
GPS_Counter equ		Page1_B+75			; loop ctl temp - use in Page 1
GPS_Comma_Cnt equ	Page1_B+76			; GPS sentence comma counter
GPS_Temp	equ		Page1_B+77			; General Temp worker
GPS_Temp_1	equ		Page1_B+78			; "                 "
GPS_Leds	equ		Page1_B+79			; Sentence quality indicators
;
; Equates for Bank 1 Flags
;
GPS_Get_Sentence	equ		0			; allow new sentence input
GPS_ISR_Ignore		equ		1			; disregard RX ISR input, processing current sentence
GPS_Find_New_Sen	equ		2			; arm "find another new sentence"
GPS_Have_Sentence	equ		3			; complete sentence is available
GPS_At_Checksum		equ		4			; at end-of-sentence checksum marker
;*************************************************************************************
;					END Of Bank 1
; ------------------------------------------------------------------------------------
; BANK 2 SRAM 		Following in Bank 2 SRAM (math)
; ------------------------------------------------------------------------------------
			org		0x120
Page2_B		equ		0x120
; **********************************************
MALat		equ		Page2_B+0			; Head of 8 doubles for moving average LAT
MALon		equ		Page2_B+16			; Head of the MA LON buffer, 8 doubles
;
MA_Ptr_H_Lat	equ		Page2_B+32			; Head Ptr Lat
MA_Ptr_E_Lat	equ		Page2_B+33			; End Ptr  Lat
MA_Ptr_W_Lat	equ		Page2_B+34			; Working Ptr Lat
;
MA_Ptr_H_Lon	equ		Page2_B+35			; Head Ptr Lon
MA_Ptr_E_Lon	equ		Page2_B+36			; End Ptr  Lon
MA_Ptr_W_Lon	equ		Page2_B+37			; Working Ptr Lon
;
; Now the actual RUNNING Ptrs, one of the sets above
;
MA_Ptr_H	equ			Page2_B+38			; Running Head
MA_Ptr_E	equ			Page2_B+39			; Running Ender
MA_Ptr_W	equ			Page2_B+40			; Running worker
MA_Ptr_W_S	equ			Page2_B+41			; Context saved working ptr
;
; Triple Precision accumulator
;
Trip2		equ		Page2_B+42			; MS
Trip1		equ		Page2_B+43			; middle
Trip0		equ		Page2_B+44			; LS
;
M_Temp_0 	equ		Page2_B+45			; worker for math routines
M_Temp_1	equ		Page2_B+46
Add_FSR_Sav equ		Page2_B+47			; FSR temp
;
; LAT / LON transfer to NAV Board holding area & ptrs.
;
Lat_Hold	equ		Page2_B+48			; need 8 slots
Lon_Hold	equ		Page2_B+56			; need 8 slots (LON)
Lat_Hold_Ptr equ    Page2_B+64			; LAT ptr
Lat_Hold_Wrk equ	Page2_B+65			; running 
Lon_Hold_Ptr equ	Page2_B+66
Lon_Hold_Wrk equ	Page2_B+67

;********************************************************************************
;					END Of Bank 2
; *******************************************************************************
;                 COMMON To ALL Banks
; *******************************************************************************
; Following at SRAM Location 0x70 appears in ALL Banks
; The Unions below are intentional - not to be changed!
; UNDERSTAND what you are changing below before you do it!!!
; These unions allow output from one conversion or math routine to serve as the
; input to the next logical routine.
 
			org		0x70
Common_B	equ		0x70

GPS_Char	equ		Common_B
GPS_Fix_Quality equ Common_B+1
Num_To_Move	equ		Common_B+2			; buffer xfer count
LCD_Move	equ		Common_B+2			; also used in LCD extraction routines
W_Temp		equ		Common_B+3			; W Save MUST appear in ALL Banks!
LCD_Shadow	equ		Common_B+4			; LCD shadow on PORTB 0-2
;
; Follows for the big binary<->decmial converters
;
;   TEM0, TEM1, TEM2, TEM3, TEMPH and TEMPL 
;
TEM0		equ		Common_B+5			; LSD INPUT/Out to  5 digit BIN->BCD cvt
D0			equ		Common_B+5			; BCD (LSB) INPUT to BCD->Binary
;			
TEM1		equ		Common_B+6			; Output from BIN->BCD
D1			equ		Common_B+6			; BCD (next LSB) INPUT to BCD->Binary

TEM2		equ		Common_B+7			; MSD INPUT/Out to  5 digit BIN->BCD cvt
D2			equ		Common_B+7			; BCD(MIDDLE) INPUT to BCD->Binary

;
TEM3		equ		Common_B+8			; Next OUTPUT to  5 digit BIN->BCD cvt
D3			equ		Common_B+8			; BCD(2nd MSB) INPUT to BCD->Binary

;
TEMPH		equ		Common_B+9			; MSD(5) OUTPUT from Binary->BCD
D4			equ		Common_B+9			; BCD(MSB) INPUT to BCD->Binary
;
TEMPL		equ		Common_B+10			; SCRATCH for Binary->BCD
;
NUMLO		equ		Common_B+11			; LSD Output of 5 digit BCD->Binary cvt
										; which is INPUT to Moving Avg routines
;
NUMHI		equ		Common_B+12			; MSD Output of 5 digit BCD->Binary cvt
										; which is INPUT to Moving Avg routines
Hold_Char	equ		Common_B+13			; Used by SSP OUT routine, may unionize later
;
Alarm		equ		Common_B+14			; Error processing
Alarm_Cnt	equ		Common_B+15
;

; ******************************************************************************
Xfer_To_Nav_Time	equ	4				; Time in sec. between data bursts to Nav Board
Sim_Play_Interval	equ	15				; Simulator transfer timer control (about 4 sec.)
; ******************************************************************************
;						End of Common Bank
; *******************************************************************************
		radix	hex

			org		0x0000
			clrf	PCLATH				; insure Page 0
			goto	Start
			org		0x0004
			goto	Int
; ***************************************************************************
			org		0x0100
Int			goto	Int_Run				; Required for Bootloader reset
Start		
; ***************************************************************************
;					DEBUG AREA
; ***************************************************************************

; ***************************************************************************
		ifdef	DEBUG
			endif
			goto	Start_Run			; away we go
; ***************************************************************************
; Issue Initial Welcome Msg
;
Initial_Msg
		call	Get_One					; into w
		movwf	Worker
		btfsc	Worker,7
		return							; all done
		call	LCDPUTCHAR				; Fire it to LCD
		incf	count
		goto	Initial_Msg				; till done
;
Get_One
		movfw	count
		addwf	PCL
;
Hello_Msg
		retlw	'G'
		retlw	'P'
		retlw	'S'
		retlw	' '
		retlw	'S'
		retlw	't'
		retlw	'a'
		retlw	'r'
		retlw	't'
		retlw	' '
		retlw	0x80					; negative delimiter is required

; ****************************************************************************
;
;
Start_Run	bcf		INTCON,GIE			; insure ints OFF
;
;
; Initialize PA & PB and TMR0
;
		BANKSEL	TRISA					; Bank 1
		movlw	0x06					;; PA is ALL digital here
		movwf	ADCON1					; set it up
		nop
;
		clrw							;; PORTA all OUTPUT
		movwf	TRISA
		movwf	TRISB					; all PORT B is output (LCD Data Bus)
;
		bsf		OPTION_REG, NOT_RBPU  	; Disable PORTB weak pull-ups
		bcf		OPTION_REG,T0CS			; TMR0 -> timer mode ARMED
		bcf		OPTION_REG,PSA			; TMR0 gets the prescaler module
	    bsf		OPTION_REG,PS0			; Set
		bsf		OPTION_REG,PS1			;  %16
		bcf		OPTION_REG,PS2			;    prescaler value.
;
; **********************************************************************
; Setup for Serial I/O under interrupts
; **********************************************************************
; PORTC used by UART I/O !
; Initialize PORTC for Serial RX/TX RC7 IN, RC6 OUT
; RC7 IN,RC6 OUT (Serial Port),RC5-OUT, RC4-IN, RC3-OUT , RC2-0 IN 


		BANKSEL	TRISC
		movlw	0x97					; configure as above~~
		movwf	TRISC					; done!
;
; Now Initialize the UART for rate,8,N,1
;
		movlw	DIVIDER					; 32D = 9600 baud w/20mHz clock
		movwf	SPBRG					; load it
		bcf		TXSTA,SYNC				; ASYNC mode
		bcf		TXSTA,BRGH				; Use low precision BRG
;
		BANKSEL	RCSTA					; BANK 0
		bsf		RCSTA,SPEN				; async mode
		bsf		RCSTA,CREN				; enable async RX

		BANKSEL	PIE1				
		bsf		PIE1,TXIE				; allow TX ints (BANK 1)
		bsf		PIE1,RCIE				; allow RX ints
;		bsf		PIE1,TMR1IE				; also allow TMR1 int
;
; Allow TMR0 interrupt - Primarily used for the Heartbeat LED on PA
;						 and general timeouts (debounce and Simulator)
		bsf		INTCON,T0IE				; unmask it, exists in all Banks
;
		BANKSEL	PORTA					; Bank 0
		clrf	Cntrlo					; set heartbeat LED flash rate
		movlw	TMR0_Rate
		movwf	Cntrhi

		clrf	A_Shadow				; LED control shadow, LED initially OFF
;
; Enable Async TX
;
		BANKSEL	TXSTA					; in BANK 1
		bsf		TXSTA,TXEN
		BANKSEL	PORTA					; back to Bank 0

		bsf		INTCON,PEIE				; arm peripheral ints

		movlw	'S'						; = ' '
		movwf	TXREG					; prime the TX one time
;
;
; Set up the SPI Port - GPS -> NAV synchronous link w/handshake

		BANKSEL TRISC 			; BANK 1
		movlw 0x00 				; Mode 1,1 SPI, middle of output (CKE = 0 so SSbar NOT needed)~
		movwf SSPSTAT 			; time sampling
		BANKSEL SSPCON 			; BANK 0
		movlw 0x32 				; Mode 1,1 SPI Master Mode, 1/64 Tosc bit time
		movwf SSPCON 			; SSP is on NOW, we are in Bank 0
; ******************************************************************************************
;
; Insure LCD Board LEDs are all OFF initially, set Shadow registers
;
		clrf	C_Shadow				; BUTTON -shadow register all clear
		clrf	C_Debounce				; BUTTON -insure button debounce tmr clr
		clrf	LCD_Shadow				; insure shadow bits clear (LEDs all OFF)
		clrf	Panel_Buts				; all Panel Button indicators OFF
;
;
; Now establish application run-time string pointers
;
		Bank_1							; Bank 1
		movlw	GPS_Raw					; a(input buffer)(Sentence Buffer)
		movwf	GPS_S_Ptr				; establish ptr
		movwf	GPS_W_Ptr				; and the running ptr
		clrf	GPS_Flags				; start COLD!
		movlw	'0'						; indicate NO fix
		movwf	GPS_Fix_Quality
		Bank_0							; back to Bank 0
;
		movlw	TX_String				; A(TX Buffer)
		movwf	TX_S_Ptr				; initial pointer TX
		movwf	TX_Work_Ptr				; and it working ptr
		clrf	TX_String
		clrf	TX_String+1
		clrf	GPS_Leds				; all indicator LEDs off
		clrf	Alarm					; DEBUG error flags
		clrf	MA_Ready				; MA available
		bcf		PORTA,5					; DEBUG~~
;
		clrf	Five_Sec				; GPS sentence tick counter, mod 5 sec (Bank 0)
;
; Initialize Pointers and set-up the MATH REGISTERS in Bank 2
;
		Bank_2							; Bank 2

		movlw	MALat					; Moving Avg buffer (LAT)
		movwf	MA_Ptr_H_Lat			; Head
		movwf	MA_Ptr_W_Lat			; and to working.

		if		NOMA	
		addlw	2						; end+1 for NO MA CALCS
		else
		addlw	d'16'					; point MA buffer end + 1
		endif

		movwf	MA_Ptr_E_Lat			; make this the ender.
;Now LON Set-Up
		movlw	MALon					; Moving Avg buffer (LON)
		movwf	MA_Ptr_H_Lon			; Head
		movwf	MA_Ptr_W_Lon			; and to working.

		if		NOMA
		addlw	2						; end+1 for NO MA CALCS
		else
		addlw	d'16'					; point MA buffer end + 1
		endif

		movwf	MA_Ptr_E_Lon			; make this the ender.
;
; Now Initialize the Transfer Pointers and prime with synchronization data
;
		movlw	Lat_Hold				; a(LAT buffer)
		movwf	Lat_Hold_Ptr
		movlw	Lon_Hold				; a(LON buffer)
		movwf	Lon_Hold_Ptr
;
		movlw	'0'						; prime transfer bufs with character '0' (zero)
		movwf	Hold_Char
		movlw	Lat_Hold				; Lat and Lon buffers are consecutive
		movwf	M_Temp_1
		call	Purge_MA				; precharge both transfer bufs
		movlw	0x0d
		movwf	Lat_Hold+6
		movlw	'Q'
		movwf	Lat_Hold+7				; for 1st pass to the Nav Board

;
		clrf	Hold_Char				; purge value is 0x00
		call	Purge_Em				; Insure MA buffs clear, returns to BANK 0
;
; ----------------------------------------------------------------------------------
; Now initialize the LCD - one time shot only-while INTS are OFF !
;
;		goto	noinit      			;~~~  remove for DEBUG on Sim
		call	LCDINIT					; done!
		movlw	LCD_LINE1
		call	LCDSDDA					; Position cursor leftmost on first line
		movlw	HIGH Hello_Msg
		movwf	PCLATH
		clrf	count
		call	Initial_Msg				; Welcome To All navigators!
noinit
; ----------------------------------------------------------------------------------
; 
; Now allow GPS input processing by setting GPS Mode !
;
		bsf		STATUS,RP0				; Bank 1
		bsf		GPS_Flags,GPS_Get_Sentence	; Enter GPS mode
		bcf		STATUS,RP0				; Bank 0

		bsf		INTCON,GIE				; allow global ints!
		bcf		LCD_Shadow,Red_T_Led	; Indicate NO SYNC initially
		clrf	Line_Ctl				; used by Simulator "Play" cmd
;
;*********************************************************************************
;					MAIN FOREGROUND
;*********************************************************************************
Main
		movfw	LCD_Shadow				; LED bits
		iorwf	Alarm,w					; include Alarm bits
		movwf	PORTB					; out to the indicator LEDS on PortB
;
; If the Alarm Bit is GPS Overflow, proceed as that was fixed in the RX ISR - indicate only!
;
		movfw	Alarm					; check problems
		andlw	0xBF					; but ignore GPS Overflow
		skpz							; skip if clear
		goto	Main					; problem, lock up with LEDS firing
;
; Access NAV Board after user button, then on every fifth GPS Sentence ( ~ every 5 seconds)
; NOTE: Probably will decrease to 3 second interval.
;
		btfss	MA_Ready,U_But_S2		; user button request for initial sync?
		goto	No_Sync_Req				; no
		btfss	Five_Sec,U_But_S2		; OK to burst?
		goto	No_Sync_Req				; not yet, but if OK above flag reset in Chk_SSI !
		call	Chk_SSI					; check SSI to synchronize boards and xfer if needed
		call	Xfer_To_Nav				; pass MA results to the Navigation Board~~

No_Sync_Req
		nop
		call	Buttons_And_Leds		; check for LCD panel button push
		clrf	Hold_Char				; in case of purge request = 0x00
		movfw	Panel_Buts				; check for a button
		sublw	0x30					; button 0 (leftmost)?
		skpnz							; skip not
		call	Purge_Em				; yes, purge MA buffer(s)~
		movfw	Panel_Buts				; grab flags again
		sublw	0x31					; boards sync user request?
		skpnz							; skip not
		bsf		MA_Ready,U_But_S2		; allow a sync process
		
		movfw	Panel_Buts				; check
		clrf	Panel_Buts				; (done with them now)
		sublw	0x32					; for Simulator Mode
		skpnz							; no, continue running live
		goto	Simulator				; enter Simulator
		bsf		INTCON,GIE

		Bank_1							; Page 1
		btfsc	GPS_Flags,GPS_Have_Sentence	; have a complete GPS sentence in?
		call	GPS_Process				; yes, go process sentence
		Bank_0							; Page 0

		bsf		INTCON,GIE				; global ints re-allowed

		goto	Main					; cycle again
; **********************************************************************
; Purge (reset) the Moving Average buffer (s) and precharge Transfer
; **********************************************************************

Purge_Em
		Bank_2
		movlw	MALat					; LAT MA buffer
		movwf	M_Temp_1
		call	Purge_MA
		movlw	MALon					; LON MA buffer
		movwf	M_Temp_1
		call	Purge_MA				; both MA buffers clear
		Bank_0							; RETURN in BANK 0 !!
		return

Purge_MA
		nop
		movfw	M_Temp_1				; head of desired buffer
		movwf	FSR
		movlw	D'16'					; num to clr
		movwf	M_Temp_0
MA_Clr	movfw	Hold_Char
		movwf	INDF
		incf	FSR,f
		decfsz	M_Temp_0
		goto	MA_Clr
		return
;************************************************************************
Chk_SSI
		bcf		Five_Sec,U_But_S2		; so only once /desired cycle time

		btfsc	LCD_Shadow,Red_T_Led	; sync previously achieved??
		goto	SSI_Sync				; yes, process SPI commands state machine
;
; Sync to the Nav Board - to be here user sync request button AND Five Second
; GPS flags must be set.  Reset the Five Second flag here for just this pass
; Can't sync if GPS is not valid.
;
		movlw	' '				; dummy to synchronize with the slave
		movwf	Hold_Char
		call	Send_SSI		; do it

		movlw	'Q'				; now query the slave
		movwf	Hold_Char
		call	Send_SSI		; NEXT return is valid

		movlw	'R'				; command slave to respond to the query
		movwf	Hold_Char
		call	Send_SSI
;
; Check for valid sync response and indicate with LED lock if so
;
		sublw	'+'				; Sync response?
		skpnz					; skip yes
		goto	SSI_Sync		; process
;		bcf		PORTA,5			; DEBUG ~~
		return					; no, just return
SSI_Sync
		bsf		LCD_Shadow,Red_T_Led	; so indicate, enter state machine
		nop
		return
;
; ***********************************************************************
;						Xfer_To_Nav
; Pass Transfer Buffers to the Navigation Board for processing
; ***********************************************************************
Xfer_To_Nav
; LAT
;		bsf		PORTA,5			; DEBUG ~~
		movlw	d'7'			; total transfer count -1 (6 data, CR, 'Q'
		movwf	ST_Wrk0			; (Bank 0)

		Bank_2
		movfw	Lat_Hold_Ptr	; a(LAT buffer)
		movwf	FSR
		movfw	INDF			; FIRST char out gets no valid response
		Bank_0
		movwf	Hold_Char		; for the physical xfer function
		movwf	ST_Wrk1			; remember this char for return verification
		call	Send_SSI		; fire it
;		movwf	TXREG			; DEBUG~~
		incf	FSR				; point to second char
		call	XTN_0			; output remaining field elements for LAT
		skpz					; if Z set, OK so far, continue with LON MA
		return					; ERRORS, just return
;
; LON MA
;
		movlw	d'7'			; total transfer count -1 (6 data, CR, 'Q'
		movwf	ST_Wrk0			; (Bank 0)

		Bank_2
		movfw	Lon_Hold_Ptr	; a(LON buffer)
		movwf	FSR
		movfw	INDF			; FIRST char out gets no valid response
		Bank_0
		movwf	Hold_Char		; for the physical xfer function
		movwf	ST_Wrk1			; remember this char for return verification
		call	Send_SSI		; fire it
;		movwf	TXREG			; DEBUG~~
		incf	FSR				; point to second char
		call	XTN_0			; output remaining field elements for LON
;		bcf		PORTA,5			; DEBUG~~
		return					; Done, if errors, they have been indicated for FG.
; ***********************************************************************************
;
; Now Do The Rest - Common for both Lat and Lon MA fields
;
XTN_0
		Bank_2
		movfw	INDF			; get next in line
		Bank_0
		movwf	Hold_Char
		movwf	ST_Wrk2			; save for test
		call	Send_SSI		; fire, on return W should have previous as response
;		movwf	TXREG			; DEBUG~~
		subwf	ST_Wrk1			; test against previous
		skpz					; skip if all ok with match
		goto	XTN_Err			; problem
		movfw	ST_Wrk2
		movwf	ST_Wrk1			; for next test
		incf	FSR				; point next
		decfsz	ST_Wrk0			; done ?
		goto	XTN_0			; no, continue
;
		bcf		PORTA,5			; DEBUG~~
		setz					; Z return indicates no errors
		return
XTN_Err
		bsf		Alarm,Lost_Sync	; indicate error to Main
		bcf		LCD_Shadow,Red_T_Led ; ... and loss of sync
		clrz					; Z clear indicates error, ruturn w/errors indicated
		return				

;
; *****************************************************************************
;						OUTPUT to the SSP
;
; Output a character to the SSP module (previously established) in Master Mode.
; INPUT  : The character in "Hold_Char"
; RETURNS: The synchronous character from the slave, in W .
;          Included transfer delay for ease of debug, will leave at some point!!
; *****************************************************************************
Send_SSI
	
;		bcf		PORTA,2				; arm slave to try other modes ~~~
		movfw	Hold_Char			; get new output
		bcf		SSPCON,WCOL
		movwf 	SSPBUF 				; put in SSPBUF
		BANKSEL SSPSTAT 			; BANK 1
Char1	 btfss 	SSPSTAT,BF 			; Data transfer complete? (Buffer Full?)
		goto 	Char1 				; if not, check again
;	
		BANKSEL	SSPBUF 				; BANK0
		movfw 	SSPBUF	 			; Get Data from SSPBUF
		movwf	Temp				; for return
;
;		bsf		PORTA,2				; disarm slave if trying other modes ~~~~~
		call	Delay_Get		; issue delay 
		movfw	Temp			; synchronous return
		return
;					Return Char in w reg (or just delay)
Delay_Get
;      Update_Test_Counter - this will disappear with a Tmr 0 counter

Delay 	movlw 	0x60 				; Simple Delay loop
		movwf   Dly1
		movlw 	0xFF 			
		movwf 	Dly0
 
DlyLoop decfsz  Dly0,F 		
		goto 	DlyLoop 		
		decfsz  Dly1,F 		
		goto 	DlyLoop

		return

; ***********************************************************************

;************************************************************************
;
; ***********************************************************************
;			BUTTON and LED Control Routines
; ***********************************************************************
; NOTE: A button push brings the PORTC Bit LOW !
;
Buttons_And_Leds
		btfsc	C_Shadow,BL_In			; working on a case now?
		return							; yes, gotta go away
		movfw	PORTC					; get all bits
		andlw	0x07					; only 0-3 interesting !	~ John Board only 0-2
		xorlw	0x07					; correct their sense		~ John Board as above
		skpnz							; continue if a hit
		return							; else just leave
		movwf	C_Shadow				; to work on 'em
;
; Got button pushes, debounce the request, then lock the bit
;

		bsf		C_Shadow,BL_In			; indicate busy now!

		movlw	0x40					; debounce value
		movwf	C_Debounce				; for the TMR0 ISR
Wait_DB
		movfw	C_Debounce				; wait
		skpz							; till zero'd out
		goto	Wait_DB					; not yet
;
; PROCESS Button hit by outputting ASCII code referencing button
;
		bcf		INTCON,GIE				; ints off now, don't need 'em
		movlw	0x04
		movwf	Worker
		movlw	0x30					; Initial "code" for external use
 
But_Chk
		rrf		C_Shadow,f				; bit to check
		skpnc							; skip if not
		goto	But_Hit					; got it, process
		addlw	1						; next code in sequence
		decfsz	Worker					; done all 4?
		goto	But_Chk					; no, continue
But_Don
		clrf	C_Shadow				; clean up
		return							; bye!
But_Hit
		bsf		C_Shadow,BL_In			; reassert the lockout
		movwf	Panel_Buts				; mark code for other button users
		movlw	0x40					; for the ISR
		movwf	C_Debounce
		bsf		INTCON,GIE				; Ints back on
Wait_DB_0
		movfw	C_Debounce
		skpz
		goto	Wait_DB_0
		clrf	C_Shadow
		return
;                             ISR
; *******************************************************************************************
; Entire context saved as using multiple banks!
; Display and shadow registers updated here.

Int_Run
		movwf 	W_Temp 			; W to TEMP register, could be bank one or zero
		swapf 	STATUS,w 		;Swap status to be saved into W
		clrf 	STATUS 			;bank 0, regardless of current bank, Clears IRP,RP1,RP0
		movwf 	S_Temp 			;Save status to bank zero STATUS_TEMP register
		movfw 	PCLATH	 		;Only required if using pages 1, 2 and/or 3
		movwf 	P_Temp 			;Save PCLATH into W
		clrf 	PCLATH 			;Page zero, regardless of current page
		bcf 	STATUS, IRP 	;Return to Bank 0
		movfw 	FSR 			;Copy FSR to W
		movwf 	F_Temp 			;Copy FSR from W to FSR_TEMP

; Now determine WHO interrupted - in PIR always check for the TX buffer
; flag LAST as it is not software cleared while transmitter active!
;
		btfsc	INTCON,T0IF				; our timer 0 ?
		goto	Tmr0ISR					; yes, visit ISR
		btfsc	PIR1,TMR1IF				; no, TMR1?
		goto	Intdie					; yup, process IF TIMER1 exists, else DIE
		btfsc	PIR1,RCIF				; UART RX?
		goto	RxISR					; got a char!

; Following test MUST be last in the chain!

		btfsc	PIR1,TXIF				; no, TX buffer int?
		goto	TxISR					; process it

;
; Unexpected interrupt, die here for debug - lock ALL LED ON for indication
;
Intdie									; BEING HERE IS VERY BAD THING
		bsf		PORTB,Yel_Led	  		; Yellow LED on LCD Board
		bsf		PORTB,Red_M_Led 		; Middle Red LED
		bsf		PORTB,Red_T_Led 		; Top RED LED

		goto	Intdie
; -----------------------------------------------------------------------------------------
;--------------------------------- TMR0 ISR -----------------------------------------------
; Primarily for debounce delays, counts, and LED heartbeat, general timeouts
;------------------------------------------------------------------------------------------
Tmr0ISR

		decf	C_Debounce				; for Buttons routines debounce
		incfsz	Cntrlo					; if overflowed, toggle time
		goto	Tmr0_xit				; not yet, exit ISR
		incfsz	Cntrhi
		goto	Tmr0_xit
;
; Got a "countdown", toggle PORTA LED status via Shadow register (Heartbeat)
; Increment and test the counter used by Simulator "Play" command
;
		btfss	Line_Ctl,0				; Simulator active?
		goto	Tmr0_0					; no, continue
		incf	Sim_Sec					; ~ 5 sec. counter
		movlw	Sim_Play_Interval
		subwf	Sim_Sec,w				; there yet?
		skpnz
		bsf		Line_Ctl,1				; yes, flag for simulator
;
Tmr0_0
		movlw	0x01
		xorwf	A_Shadow,f				; toggle the LED bit
		btfsc	A_Shadow,0				; Turn on?
		goto	Toff					; already on, turn off
		bsf		LCD_Shadow,Heart_Led 	; was off, turn ON
		nop
		goto	T0						; exit
Toff
		bcf		LCD_Shadow,Heart_Led 	; was on, turn OFF
		nop

; Set FIX LED based upon GPS_Leds from foreground
; If set, leave it on, if clear kill the FIX indicator

		btfss	GPS_Leds,1
		goto	F_Off					; turn off
		bsf		LCD_Shadow,Yel_Led		; turn on or leave on
		goto	T0						; bye
F_Off
		bcf		LCD_Shadow,Yel_Led
T0
		clrf	Cntrlo
		movlw	TMR0_Rate				; reset heartbeat rate
		movwf	Cntrhi
;
Tmr0_xit
		bcf		INTCON,T0IF				; dispatch this int
		clrf	TMR0					; clear the counter and ...

;
; Common context restore and return from interrupt

ISR_Xit
		movfw 	F_Temp					; Restore ...
		movwf	FSR						; FSR
		movfw 	P_Temp 					;Restore PCLATH
		movwf 	PCLATH 					;
		swapf 	S_Temp,w 				;Swap STAT
		movwf 	STATUS 					;Move W into STATUS register
		swapf 	W_Temp,f 				;Swap W_TEMP
		swapf 	W_Temp,w 				;Swap W_TEMP into W
		retfie							; see ya!
; **********************************************************************
; RX ISR - accumulate a complete GPS sentence and set appropriate flags
;        - unused characters simply discarded
;		 - a bad sentence-in-progress is purged - there will be more
;		 - validate sentence checksum before allowing NAV processing
; **********************************************************************
RxISR
		movfw	RCREG					; grab the char
		movwf	GPS_Char				; stash in global SRAM
		bsf		STATUS,RP0				; Bank 1
		btfss	GPS_Flags,GPS_Get_Sentence	; GPS MODE input requested???(This set in FG)
		goto	GPS_Purge				; no, purge input char and exit
										; yes, continue (in Page 1 !) 
;
; In GPS mode, discard chars until the '$' is found, indicating start of a sentence.
; When found, mark it and continue reading.  WE ARE IN BANK 1 NOW!!
;
Rx_GPS
		btfsc	GPS_Flags,GPS_ISR_Ignore	; GPS ISR read complete?
		goto	GPS_Purge				; if set, purge incoming char, sentence complete
;
		movfw	GPS_W_Ptr				; to the working ptr
		movwf	FSR						; and INDF
;
		movfw	GPS_Char				; grab incoming char
		movwf	INDF					; stash it to GPS buffer

		btfsc	GPS_Flags,GPS_Find_New_Sen	; already working a sentence??
		goto	Rx_In_Sentence			; yes, proceed
		sublw	'$'						; no, start of sentence?
		skpz							; skip if YES
		goto	Rx_GPS_1				; no, just leave
;
; Got the start character '$'
;
		bsf		GPS_Flags,GPS_Find_New_Sen	; mark it as so
		clrf	TEM0					; reset checksum target
		Bank_0
		clrf	Alarm_Cnt				; start a running counter for buf overflow test 
		incf	Five_Sec				; sentences occur about once/sec, indicate got one
		movfw	Five_Sec
		andlw	0x07					; only consider bits 0-2
		xorlw	Xfer_To_Nav_Time		; this establishes the burst interval
		skpz							; skip when there
		goto	Rx_In_Sentence			; not yet

		clrf	Five_Sec				; there, reset counter
		bsf		Five_Sec,U_But_S2		; FG OK for the  burst to NAV Board
										; (FG resets bit after each burst).
		movlw	0x02					; to
		xorwf	A_Shadow,f				; toggle the burst LED bit
		btfsc	A_Shadow,1				; Turn on  ?
		goto	Toff_1					; already on, turn off
		bsf		LCD_Shadow,Five_Sec_Led	; was off, turn ON
		nop
		goto	Rx_In_Sentence			; exit this
Toff_1
		bcf		LCD_Shadow,Five_Sec_Led	; was on, turn OFF
;
; Process characters within live sentence
;
Rx_In_Sentence
		Bank_1
		movfw	INDF					; most recent char in buffer
		call	Checksum				; update running checksum
		incf	Alarm_Cnt				; running cntr
		movwf	GPS_Recent				; for ending test		
		incf	FSR
		movfw	FSR
		movwf	GPS_W_Ptr				; point next slot
;
; Determine if complete sentence in, GPS enders always CR+LF
;
		movlw	0x0A					; LF = ender
		subwf	GPS_Recent,w			; got it?
		skpz							; skip if yes
		goto	Rx_GPS_1				; no, just exit
;
; Verify Checksum OK
;
		movlw	0x04					; back-up FSR...
		subwf	FSR,f					; to point at ms checksum character
;
; Cvt send CS -> Binary and check against running CS
;
		call	Cvt_Checksum			; convert sent CS -> hex in w
		subwf	TEM0,w					; need zero result to accept sentence
		skpnz
		goto	AG						; CS OK, proceed
		bsf		Alarm,CS_Error			; CS FAILURE, flag and discard sentence
		goto	A_Bad_Sentence_CS		; non-fatal error
;
; All Good
;
AG
		bsf		GPS_Flags,GPS_Have_Sentence	; yes, mark ender LF found!
		bsf		GPS_Flags,GPS_ISR_Ignore	; ... and ISR read complete.
		bcf		STATUS,RP0				; Bank 0
		bsf		LCD_Shadow,Red_M_Led	; light Debug LED (complete sentence is available
		goto	GPS_Purge				; exit

;
Rx_GPS_1
; Debug check for buffer overflow 

		movlw	D'70'					; max size of GPS buffer 
		subwf	Alarm_Cnt,w				; check there ... 
		skpz							; not zero is GOOD
		goto	GPS_Purge				; so just continue normally (leave clean)				

; Input buffer attempted to overflow - kill this sentence and restart input process
A_Bad_Sentence
		bsf		Alarm,GPS_Overflow		; yikes, indicate overflow, kill overflow sentence.
A_Bad_Sentence_CS						; CS error already indicated, now abort sentence
		clrf	GPS_Flags				; reset ISR
		bsf		GPS_Flags,GPS_Get_Sentence	; reallow input of new sentence
		movfw	GPS_S_Ptr				; reset GPS buffer input
		movwf	GPS_W_Ptr				; pointers.
;
GPS_Purge
		bcf		STATUS,RP0				; back to Bank 0
		goto	ISR_Xit					; exit ISR
; ************************************************************************
; ************************************************************************
; Process the sentence checksum most recent in W and (INDF)
; Do not process the '$', '*', CR, or LF
; Strip and merge this routine into "within sentence" processing someday!
; Being here means the leading '$' has already been recognized, so we are
; processing upto, but not including, the '*', indicating ASCII representation
; of the checksum immediately follows.
;
Checksum
		sublw	'$'						; at start flag
		skpnz							; no, continue
		goto	Check_Out				; yup, no process this one!
		btfsc	GPS_Flags,GPS_At_Checksum
		goto	Check_Out				; ignore any char following the cs marker
		movfw	INDF
		sublw	'*'						; cs marker?
		skpz							; yes
		goto	CScc					; no, continue, process this char.
		bsf		GPS_Flags,GPS_At_Checksum
		goto	Check_Out				; yes, mark and exit
CScc
		movfw	INDF					; process -> running CS
		xorwf	TEM0,f					; partial sum
Check_Out
		movfw	INDF					; xit 
		return							; with char intact
; ************************************************************************
; ASCII->Hex conversion, used only by Checksum calculator
;   - uses TEM1 in common and in-place operations in the sentence
;   - this routine used only for CS, as input params narrowly defined!
;   - ASCII 0-9 and A-F processed
; ************************************************************************
Chk_Cvt_1
		btfsc	TEM1,6					; ASCII 0-9 or A-F test
		goto	Cvt_Let					; process A-F
		movfw	TEM1
		andlw	0x0F					; strip zone
Cvt_R	movwf	TEM1					; common exit
		return
Cvt_Let
		movlw	d'55'					; letter converter
		subwf	TEM1,w
		movwf	TEM1
		return
; ************************************************************************
; Cvt_Checksum - process the two ASCII checksum characters representing
; 				 the actual calculated checksum.
; ************************************************************************
Cvt_Checksum
		movfw	INDF					; 1st char following the "*"
		movwf	TEM1					; working area in common
		call	Chk_Cvt_1				; convert ms character
		swapf	TEM1,w					; align
		movwf	INDF					; result overwrites sentence slot
		incf	FSR						; point->ls character
		movfw	INDF					; get it
		movwf	TEM1
		call	Chk_Cvt_1				; process ls
		decf	FSR						; point back at partial result
		iorwf	INDF,w					; create calculated checksum
		return							; carry on
; ************************************************************************
; ************************************************************************
;                         TX UART ISR
; Output the string at TX_String (TX_Work_Ptr) until null encountered
; Kill the TX int enable, but leave TX armed following completed output
;
TxISR
		movfw	TX_Work_Ptr				; get current position
		movwf	FSR						; to indirect
		movfw	INDF					; output character
		skpnz							; delimiter?
		goto	TX_ISR_Off				; yup, shut down ints!
;
		incf	FSR,f					; no, bump and save pointer
		movwf	TXREG					; Fire the character 
		movfw	FSR
		movwf	TX_Work_Ptr				; save running ptr
		goto	ISR_Xit					; done this pass!
;
TX_ISR_Off
		BANKSEL	PIE1
		bcf		PIE1,TXIE				; TX int OFF now
		BANKSEL	PORTA
		goto	ISR_Xit					; done till next string

; ******************************************************************************
;						GPS PROCESSING
; ******************************************************************************
GPS_Process
		bsf		STATUS,RP0
		bcf		GPS_Flags,GPS_Get_Sentence	; disarm RX for now, stay in PAGE 1 !
;
; We have a complete sentence, for debug, move to TX buffer and output
; Enter here with INTS OFF and in PAGE 1, via CALL from Main .  We need to verify
; that sentence is a GPGGA position type, else discard it and get another.
; 
		bcf		STATUS,RP0				; PAGE 0
		movfw	TX_S_Ptr
		movwf	TX_Work_Ptr				; prime output ptrs
		movwf	Gen_Temp
		bsf		STATUS,RP0				; PAGE 1

		movfw	GPS_S_Ptr				; GPS DATA Ptr head-of-sentence
		movwf	FSR						; to extract
		movwf	GPS_FSR_Sav				; save this
		movwf	GPS_Temp_1				; ...again
		bcf		GPS_Flags,GPS_Get_Sentence	; and indicate GPS Mode is OFF to the RX
		clrf	GPS_Comma_Cnt			; sentence position
;
; Verify the GPGGA sentence ID - we are in Page 1 FSR pointing to head
;
		movlw	'$'
		subwf	INDF,w					; got the $ sign?
		skpz
		goto	Sentence_NG				; NO, fail
		incf	FSR
		movlw	'G'						; G ?
		subwf	INDF,w					; check
		skpz							; skip if YES
		goto	Sentence_NG				; NO, exit all this
		incf	FSR			
		movlw	'P'						; P ?
		subwf	INDF,w
		skpz							; skip YES
		goto	Sentence_NG
		incf	FSR
		movlw	'G'						; first G ?
		subwf	INDF,w
		skpz							; skip YES
		goto	Sentence_NG
		incf	FSR
		movlw	'G'						; second G ?
		subwf	INDF,w
		skpz							; skip YES
		goto	Sentence_NG
		incf	FSR
		movlw	'A'						; final A ?
		subwf	INDF,w				
		skpz							; skip YES
		goto	Sentence_NG
;
; We have the correct sentence.  Now slew forward six (6) commas to point to the
; quality of fix indicator. Value 1 or 2 is OK, we are locked, else NG.  Right now
; we are pointing at the 'A' .
;
		movfw	FSR						; save this
		movwf	GPS_Temp				; location
;
; Extract the UTC Time
;

		incf	FSR						; point to 1st TIME character -1
		movlw	0x06					; TIME = 6 char
		movwf	Num_To_Move
		call	GPS_Mov_To_Output		; extract TIME info to write buffer
		movfw	GPS_Temp				; restore
		movwf	FSR						; FSR for continuing extraction
		movfw	GPS_Temp_1				; restore slew pointer
		movwf	GPS_FSR_Sav				; and continue.
;
		movlw	6						; comma fields
		movwf	GPS_Comma_Cnt			; number to slew

GPS_Verify_Fix
		incf	FSR						; point char to check
		movlw	','
		subwf	INDF,w					; checking for comma
		skpz							; skip if found
		goto	GPS_Verify_Fix			; not, continue forward
		decfsz	GPS_Comma_Cnt			; at position?
		goto	GPS_Verify_Fix			; no, continue forward
;
; Now check fix indicator
;
		incf	FSR
		movfw	INDF					; get the indicator
		movwf	GPS_Fix_Quality			; save the indicator
		sublw	0x30					; if zero, not valid fix
		skpnz
		goto	Fix_NG					; fix info NOT valid
;
; WE have a VALID sentence, proceed !
;
		bcf		STATUS,RP0				; Page 0
		bsf		GPS_Leds,1				; Allow the FIX LED (ISR processes)
		bsf		STATUS,RP0				; back to Page 1
		clrf	GPS_Comma_Cnt			; reset the comma counter

;
; Extract the LAT from the msg
;
GPS_L
		movfw	GPS_FSR_Sav				; abs position in sentence
		movwf	FSR						; point to it
		movlw	','					    ; check char of GPS data for delim
		subwf	INDF,w
		skpz							; skip if a comma
		goto	GPS_L1					; not yet, slew into string
;
; Comma delimiter found
;
		incf	GPS_Comma_Cnt			; note found
		movfw	GPS_Comma_Cnt			; 
		sublw	2						; LAT follows 2nd comma
		skpnz							; not yet
		goto	GPS_L2					; at 2nd comma now, FSR bump will point, PROCESS
;	
GPS_L1
		incf	FSR						; and point next
		movfw	FSR
		movwf	GPS_FSR_Sav				; save NEXT slot position
		goto	GPS_L					; continue looking
;
; We are positioned at the LAT-1
GPS_L2
		movlw	d'9'					; LAT = 9 char
		movwf	Num_To_Move
		call	GPS_Mov_To_Output		; extract LAT info to write buffer
;
; Now extract LON info
;
		incf	GPS_FSR_Sav				; bump to
		incf	GPS_FSR_Sav				;  point at LON-1
		movfw	GPS_FSR_Sav
		movwf	FSR
		movlw	d'10'					; LON = 10 char
		movwf	Num_To_Move
		call	GPS_Mov_To_Output		; extract LON info to write buffer
;
;
; Complete the output string with Fix Quality indicator
;
		bcf		STATUS,RP0				; Now BANK 0
		movfw	Gen_Temp				; output ptr
		movwf	FSR						; to indirect
		movlw	' '
		movwf	INDF
		incf	FSR
		movfw	GPS_Fix_Quality
		movwf	INDF
		incf	FSR
		call	Finish_String			; complete the msg
;
		movfw	TX_S_Ptr				; head of the output string
		movwf	FSR						; to the FSR
		call	MA_Calc					; process lat/lon, create transfer buffers, out -> LCD
		movfw	TX_S_Ptr				; head of output to the TX
		call	Issue_Data				; TX output (if desired) and exit with FIX indicator ON !
		goto	Sentence_Xit			; see ya
;
; If the sentence was OK, issue message with quality indicator
;
Fix_NG
		bcf		STATUS,RP0				; now Page 0
		bcf		GPS_Leds,1				; FIX indicator off (ISR processes)
		movfw	TX_Work_Ptr				; TX String ptr
		movwf	FSR
		movlw	'N'						; Msg -> terminal for debug
		movwf	INDF
		incf	FSR
		movlw	'F'
		movwf	INDF
		incf	FSR	
		movlw	' '
		movwf	INDF
		incf	FSR
		movfw	GPS_Fix_Quality			; Fix quality
		movwf	INDF
		incf	FSR
		call	Finish_String			; complete the msg
;
		movlw	LCD_LINE1				; make Line 1
		call	LCDSDDA					; do it
		movlw	LCD_CLEAR				; purge it
		call	LCDPUTCMD
		movlw	'N'
		call	LCDPUTCHAR
		movlw	'F'
		call	LCDPUTCHAR				; output 'NF' to LCD
;		
		movfw	TX_S_Ptr				; head of the output string
		call	Issue_Data				; and exit!
;
Sentence_NG								; in ALL cases,
Sentence_Xit							; Clean up and exit

		bsf		STATUS,RP0				; Page 1
		movfw	GPS_S_Ptr				; reset
		movwf	GPS_W_Ptr				; pointers
		clrf	GPS_Flags				; and flag bits

		bsf		GPS_Flags,GPS_Get_Sentence	; reallow input

		bcf		STATUS,RP0				; Page 0
		bcf		LCD_Shadow,Red_M_Led	; Sentence received indicator OFF
		return							; back to the FG		
; ******************************************************************************
Issue_Data
; NOTE: After transmitting the requested string, the TX ISR kills it's
;       own interrupt enable flag so TX won't spuriously interrupt.
; INPUT: Ptr to zero-terminated output string
;
Issue_Data
		ifdef	ISSUE_RS232

		movwf	TX_Work_Ptr				; initial message.
		BANKSEL	PIE1
		bsf		PIE1,TXIE				; TX int ON now
		BANKSEL	PORTA

		endif

		return
;
;
; *****************************************************************************
;			Compute MA and	WRITE NAV INFO TO THE LCD, Create Transfer Buffers
;
; Extract and output two, 16 byte strings to the LCD
; Enter with W pointing to the head of the output buffer
; Calls conversion routines and Moving Average calculators
; Executes in Bank 0 and uses Common and Bank 0 variables except where noted.
; Also creates "Transfer Buffers" for MA results, used to pass smoothed coordinates
;     to the Navigation Board ("C" coded)
;
; *****************************************************************************

		ifdef	LCD_Main
;
; Follows executes and uses BANK 0 except where noted - don't change things
; without fully understanding where variables are located!
;
MA_Calc
		movlw	d'15'					; buffer position
		movwf	Worker					; pass to common code

		movfw	FSR
		movwf	LCD_Move				; save FSR
; Just move mm.mmmm fields
		addlw	d'9'					; first target position
		movwf	FSR						; to parse the string
;
		movlw	LCD_CLEAR				; purge it
		call	LCDPUTCMD
		movlw	LCD_LINE1				; make Line 1
		call	LCDSDDA					; do it

		call	MA_Common_1				;~~ common L/L code

		Bank_2
; 
; Prime WORKING PTRS with the LAT running ptrs
;

		movfw	MA_Ptr_W_Lat			; LAT
		movwf	MA_Ptr_W				; running working ptr

		movfw	MA_Ptr_H_Lat
		movwf	MA_Ptr_H				; Head ptr

		movfw	MA_Ptr_E_Lat			; End ptr
		movwf	MA_Ptr_E				; Now set-up good for LAT calcs	

		call	MA_Process				; execute moving average routines

		movfw	MA_Ptr_W_S				; save the working LAT context
		movwf	MA_Ptr_W_Lat			; for next cycle

;
; Now insert the Moving Average into the LCD Message
;
		movfw	Trip0					; LS to cvt
		movwf	D0
		movfw	Trip1
		movwf	D2
		Bank_0
		call	B2A						; Binary -> BCD
;
; Now move the MA to the LCD (must make BCD to ASCII by inserting the zone)
; Source in TEMPH, D0-D4, MSB first (running backwards)
; Will look like this: m.mmmm  on local LCD
;
; First, establish the Transfer Buffer target address. Much bank switching required
; to pull this off !
;
		Bank_2
		movlw	Lat_Hold+2				; ptr to fractional mmmm slot
		Bank_0
		movwf	DELAY					; save this

		movlw	' '						; Space
		call	LCDPUTCHAR				; out to LCD
		movlw	D'4'
		movwf	Temp_0					; Page 0 Temp
		movlw	TEMPH					; a(MSB)
		movwf	FSR
		movfw	INDF					; get MS
		addlw	0x30					; make ASCII
		Bank_2
		movwf	Lat_Hold				; to MS of transfer buffer
		Bank_0
		call	LCDPUTCHAR				; MS out to the local LCD
		movlw	'.'
		Bank_2
		movwf	Lat_Hold+1				; to transfer
		Bank_0
		call	LCDPUTCHAR				; delim out to local LCD
		decf	FSR
LCD_02
		movfw	INDF
		addlw	0x30					; make BCD -> ASCII (4 fractional m digits)

		movwf	Worker					; hold the character
		movfw	FSR
		movwf	COUNT					; save current FSR
		movfw	DELAY
		movwf	FSR						; Transfer slot adx (running) in Bank 2
		movfw	Worker					; the character
		Bank_2
		movwf	INDF					; -> Transfer buffer
		Bank_0
		incf	FSR
		movfw	FSR
		movwf	DELAY					; save Transfer FSR for next pass
		movfw	COUNT					; and
		movwf	FSR						; restore Bank_0 FSR
		movfw	Worker					; the character
;
		call	LCDPUTCHAR
		decf	FSR,f
		decfsz	Temp_0					; Page 0 temp
		goto	LCD_02					; till done.
;										; insert delimiter in transfer buffer
		movlw	0x0d					; = cr
		Bank_2
		movwf	Lat_Hold+6				; (For "C" on Nav Board use)
		movlw	'Q'						; the Query for Nav Board
		movwf	Lat_Hold+7
		Bank_0


; **************************************************************************
; LON - extract from text processing buffer -> LCD      LON
; **************************************************************************
Do_Lon
		movlw	D'26'					; Buffer position
		movwf	Worker					; for common code.

		movlw	LCD_LINE2				; head of Line 2
		call	LCDSDDA					; position it

		movfw	LCD_Move				; restore original FSR
; Just move mm.mmmm fields of LON
		addlw	D'20'					; first target position
		movwf	FSR						; to parse the string

		call	MA_Common_1				; common routine
;
		Bank_2
; 
; Prime WORKING PTRS with the LON running ptrs
;

		movfw	MA_Ptr_W_Lon			; LON
		movwf	MA_Ptr_W				; running working ptr

		movfw	MA_Ptr_H_Lon
		movwf	MA_Ptr_H				; Head ptr

		movfw	MA_Ptr_E_Lon			; End ptr
		movwf	MA_Ptr_E				; Now set-up good for LON calcs	

		call	MA_Process				; execute moving average routines

		movfw	MA_Ptr_W_S				; save the working LON context
		movwf	MA_Ptr_W_Lon			; for next cycle
;
; Now insert the Moving Average into the LCD Message
;
		movfw	Trip0					; LS to cvt
		movwf	D0
		movfw	Trip1
		movwf	D2
		Bank_0
		call	B2A						; Binary -> BCD
;
; Now move the MA to the LCD (must make BCD to ASCII by inserting the zone)
; Source in D0-D4, MSB first (low memory)
;
		Bank_2
		movlw	Lon_Hold+2				; ptr to fractional mmmm slot
		Bank_0
		movwf	DELAY					; save this

		movlw	' '						; Space
		call	LCDPUTCHAR				; out to LCD
		movlw	D'4'
		movwf	Temp_0					; Page 0 Temp
		movlw	TEMPH					; MSB
		movwf	FSR
		movfw	INDF					; get MS
		addlw	0x30					; make ASCII

		Bank_2
		movwf	Lon_Hold				; to MS of transfer buffer
		Bank_0
		call	LCDPUTCHAR				; MS out to the local LCD
		movlw	'.'
		Bank_2
		movwf	Lon_Hold+1				; to transfer
		Bank_0
		call	LCDPUTCHAR				; delim out
		decf	FSR
LCD_02X
		movfw	INDF
		addlw	0x30					; make BCD -> ASCII (4 fractional m digits)

		movwf	Worker					; hold the character
		movfw	FSR
		movwf	COUNT					; save current FSR
		movfw	DELAY
		movwf	FSR						; Transfer slot adx (running) in Bank 2
		movfw	Worker					; the character
		Bank_2
		movwf	INDF					; -> Transfer buffer
		Bank_0
		incf	FSR
		movfw	FSR
		movwf	DELAY					; save Transfer FSR for next pass
		movfw	COUNT					; and
		movwf	FSR						; restore Bank_0 FSR
		movfw	Worker					; the character
;
		call	LCDPUTCHAR
		decf	FSR,f
		decfsz	Temp_0					; Page 0 temp
		goto	LCD_02X					; till done.
;
		movlw	0x0d					; = cr
		Bank_2
		movwf	Lon_Hold+6				; (For "C" on Nav Board use)
		movlw	'Q'						; the Query for Nav Board
		movwf	Lon_Hold+7
		Bank_0

; *******************************************************************
Q_In
		movlw	' '
		call	LCDPUTCHAR
		movfw	LCD_Move				; restore original FSR
		addlw	D'29'					; fetch
		movwf	FSR						; and
		movfw	INDF					; insert
		call	LCDPUTCHAR				; the Quality indicator.
		return

; *********************************************************************
;					MA Common Code Segment 1
; *********************************************************************
MA_Common_1
		movlw	d'7'					; L/L count (mm.mmmm)
		movwf	Temp_0					; Page 0 temp					
LCD_01
		movfw	INDF					; grab one
		incf	FSR
		call	LCDPUTCHAR				; out to LCD
		decfsz	Temp_0					; till done
		goto	LCD_01
;
; Now pass ASCII data to the conversions for MA processing
; Insert LS first, moving BACK to LS, the Zone must be stripped to make BCD
; (Bank 1 -> Bank 2)
;
		movfw	Worker					; grab offset ...
		addwf	LCD_Move,w				; Buffer + passed offset ...
		movwf	FSR						; to FSR
;
		movfw	INDF					; LAST
		andlw	0x0F					; strip ZONE
		movwf	D0
;
		decf	FSR
		movfw	INDF					; <-
		andlw	0x0F					; strip ZONE
		movwf	D1
;
		decf	FSR
		movfw	INDF					; <-
		andlw	0x0F					; strip ZONE
		movwf	D2
;
		decf	FSR
		movfw	INDF					; <-
		andlw	0x0F					; strip ZONE
		movwf	D3
;
		decf	FSR
		decf	FSR						; <- <- (no 'period' allowed!)
		movfw	INDF
		andlw	0x0F					; strip ZONE
		movwf	D4
;
; Now convert BCD -> Binary at NUMLO and NUMHI
;
		call	dec2bin16				; convert to BCD -> NUMLO and NUMHI
		return

		endif

; *********************************************************************
;					GENERAL SOURCE TO OUTPUT EXTRACTOR
;
; CALL with FSR at -1 position in SOURCE, char count in "Num_To_Move"
; Extract the nine (9) LAT chars to the TX buffer  (xxxx.yyyy)
; Also the nine (9) LON characters, or whatever ya want from the string
; Each call here places a space delimiter following the field just moved to output
;
; Returns with SOURCE PTR pointing to next source sentence char
; Returns with OUTPUT PTR positioned for next text insertion
; Returns in BANK 1, this is where GPS sentence string is based
;
; REMEMBER: SAVE source ptr (GPS_FSR_Sav) when calling this if needed to back-up
;           within the source string.  CAUTION!
;
GPS_Mov_To_Output
		incf	FSR
GPS_L3
		movfw	INDF					; grab Page 1 char
		movwf	GPS_Char				; save in global SRAM
		incf	FSR						; point next
		movfw	FSR
		movwf	GPS_FSR_Sav				; hold for next cycle
;
		bcf		STATUS,RP0				; NOW PAGE 0
		movfw	Gen_Temp					; output ptr
		movwf	FSR						; to indirect
		incf	Gen_Temp
		movfw	GPS_Char				; get the char
		movwf	INDF					; to the output string
		bsf		STATUS,RP0				; back to PAGE 1
		movfw	GPS_FSR_Sav				; source ptr
		movwf	FSR
		decfsz	Num_To_Move
		goto	GPS_L3					; not, continue!

		bcf		STATUS,RP0				; DONE, now Page 0
		movfw	Gen_Temp
		movwf	FSR
		incf	Gen_Temp				; leave pointing to next slot
		movlw	' '						; delimiter
		movwf	INDF
		bsf		STATUS,RP0				; leave in Page 1
		return


;**********************************************************************
Finish_String
		movlw	0x0D					; CR
		movwf	INDF
		incf	FSR
		movlw	0x0A					; LF
		movwf	INDF
		incf	FSR
		clrw							; ENDER
		movwf	INDF
		return

; *********************************************************************
;							SIMULATOR
; Called on Button S3 depressed.  Allow user to input test MA values to the Transfer
; Buffers for SSI xfer to the Navigation Board.
; - entered with INTS OFF
; - Synchronization with NAV is not assumed
; - Serial I/O POLLING here to avoid hacking the ISR as Sim may become standalone
;   someday and removed from this module.

;						SIMULATOR COMMANDS
;
; s - force NAV Board sync
; x - accept RS-232 Latitude Moving Average value in form l.llll (cr)
; y - accept RS-232 Longitude MA as above
; t - transfer to the Nav Board (also forces sync)
; r - return to caller
; p - "Play" : issue 't' transfer on 5 sec. clock to simulate GPS timing
; o - turn "Play" process off
;
; *********************************************************************
;
Simulator
		Bank_0
		clrf	GPS_Flags				; Insure GPS mode disarmed

		call	LCDCLEAR				; set up LCD
		movlw	LCD_LINE1
		call	LCDSDDA					; Position cursor leftmost on first line
;
		movlw	'S'
		movwf	TX_String				; Initial msg to both TX and LCD
		call	LCDPUTCHAR
		movlw	'I'
		movwf	TX_String+1
		call	LCDPUTCHAR
		movlw	'M'
		movwf	TX_String+2
		call	LCDPUTCHAR
		movlw	D'13'
		movwf	TX_String+3
		clrw
		movwf	TX_String+4
;
		bsf		INTCON, GIE				; global ints ON
		movfw	TX_S_Ptr				; head of the output string
		call	Issue_Data				; and fire it
;
; Disarm RX Ints as we will poll for our commands
;
		BANKSEL	PIE1				
		bcf		PIE1,RCIE				; disallow RX ints
		BANKSEL	PORTA
;
; Poll for a valid Simulator command - simply disregard bogus command input.
; Then process command after parse.
;
; Insure appropriate foreground status indicators are in proper state
;
		clrf	LCD_Shadow				; global status
		clrf	Alarm					; and alarms reset
		clrf	Line_Ctl				; used by Play command
;
Sim_Cmd_In
		btfsc	PIR1,RCIF				; check for a char
		goto	Got_Sim					; have one now
		movfw	LCD_Shadow				; not yet, but display LED bits
		iorwf	Alarm,w					; include Alarm bits
		movwf	PORTB					; out to the indicator LEDS on PortB
		btfss	Line_Ctl,0				; "Play" active?
		goto	Sim_Cmd_In				; no, continue loop
		btfss	Line_Ctl,1				; yes, ready for "Transfer"?
		goto	Sim_Cmd_In				; no, continue loop
		clrf	Sim_Sec					; yes, reset ISR counter
		bcf		Line_Ctl,1				; reset ISR counter hit flag
		goto	Sim_Xfer				; ... and do a transfer
;
Got_Sim
		movfw	RCREG					; got one
		movwf	RX_Char					; hold it
;
		movlw	0x0d
		subwf	RX_Char,w
		skpnz
		goto	Sim_Cmd_In				; disregard cr in

;
		movfw	RX_Char
		movwf	TXREG					; echo input
		movlw	's'						; Start Sync?
		subwf	RX_Char,w
		skpnz							; skip no
		goto	Sim_Sync
;
		movlw	'x'						; LAT input?
		subwf	RX_Char,w
		skpnz
		goto	Sim_Lat					; do it
;
		movlw	'y'						; LON input?
		subwf	RX_Char,w
		skpnz
		goto	Sim_Lon					; do it
;
		movlw	't'						; Transfer to NAV?
		subwf	RX_Char,w
		skpnz
		goto	Sim_Xfer				; do it
;
		movlw	'r'						; return to Main?
		subwf	RX_Char,w
		skpnz
		goto	Sim_Ret					; yes!
;
		movlw	'p'						; Play ?
		subwf	RX_Char,w
		skpnz
		goto	Sim_Play				; yes!
;
		movlw	'o'						; Play off?
		subwf	RX_Char,w
		skpnz
		goto	Sim_Play_Off			; yes!
;
; Unexpected Input, retry
;
		movlw	'?'
		call	Poll_Push
		movlw	D'13'
		call	Poll_Push
		goto	Sim_Cmd_In				; try again
;
; *********************************************************************
; Command Processing Routines follow
; *********************************************************************
; Play - allow transfer operation on ~5 sec. interval to simulate GPS timing
; to the Nav Board. TMR0 ISR increments Sim_Sec when active.
Sim_Play
		clrf	Sim_Sec					; reset 5 sec. counter
		bsf		Line_Ctl,0				; indicate "Play" active
		goto	Sim_Happy				; and exit
;
Sim_Play_Off
		bcf		Line_Ctl,0				; indicate "Play" inactive
		goto	Sim_Happy
;
; *******************************************************************************
Sim_Sync
;
; Prep the transfer buffers
;
		Bank_2
		movfw	Lat_Hold_Ptr			; a(LAT)
		addlw	D'6'					; boilerplate start adx
		movwf	FSR
		movlw	D'13'					; Boilerplate 1
		movwf	INDF
		incf	FSR
		movlw	'Q'						; Boilerplate 2
		movwf	INDF
;
		movfw	Lon_Hold_Ptr			; a(LON)
		addlw	D'6'					; boilerplate start adx
		movwf	FSR
		movlw	D'13'					; Boilerplate 1
		movwf	INDF
		incf	FSR
		movlw	'Q'						; Boilerplate 2
		movwf	INDF
;
		Bank_0
		bsf		MA_Ready,U_But_S2		; allow a sync process when necessary
		bsf		Five_Sec,U_But_S2		; allow entry to the sync master!

		call	Chk_SSI					; (s) check SSI to synchronize boards and xfer if needed
		call	Xfer_To_Nav				; pass MA results to the Navigation Board~~
		nop								; BOTH GPS and NAV synch LEDS should be red now!
		goto	Sim_Happy				; issue happy '!', get next cmd
; *********************************************************************
; Sim_Lat
; *********************************************************************
Sim_Lat
		Bank_2
		movfw	Lat_Hold_Ptr			; a(LAT) Transfer Buffer
		goto	Sim_Trans_Move			; enter common
Sim_Lon
		Bank_2
		movfw	Lon_Hold_Ptr			; a(LON) Transfer Buffer, and fall into next	

; *********************************************************************
; Common routine to prime Lat and Lon Transfer buffers
; *********************************************************************

Sim_Trans_Move							; (x) Accept a test LAT MA Value ( l.llll) -> Transfer
		nop								; enter with w<- appropriate transfer buffer adx!
		movwf	FSR
		Bank_0
		call	Poll_Get				; grab first
		Bank_2
		movwf	INDF					; -> Transfer
		incf	FSR
		movlw	'.'						; the '.' -> Transfer
		movwf	INDF
		incf	FSR
		Bank_0
		movwf	TXREG					; echo the '.'
;
; Now the remaining 4 fractional digits and the boilerplate
;
		movlw	D'4'
		movwf	Worker
Sim_4_In
		call	Poll_Get				; fetch input
		Bank_2
		movwf	INDF
		incf	FSR
		Bank_0
		decfsz	Worker
		goto	Sim_4_In				; till fraction is in
		movlw	D'13'					; boilerplate in
		Bank_2
		movwf	INDF
		incf	FSR
		movlw	'Q'
		movwf	INDF					; done now
		Bank_0
		goto	Sim_Happy				; and leave
; ********************************************************************
Sim_Xfer								; transfer the buffers to NAV Board
		bsf		MA_Ready,U_But_S2		; allow a sync process when necessary
		bsf		Five_Sec,U_But_S2		; allow entry to the sync master!

		call	Chk_SSI
		call	Xfer_To_Nav				; transfer L/L
		goto	Sim_Happy				; and leave
; ********************************************************************		
Sim_Ret
		BANKSEL	PIE1				
		bsf		PIE1,RCIE				; reallow RX ints
		BANKSEL	PORTA

		goto	Main					; back to the wars!
; *********************************************************************
Sim_Happy
		movlw	'!'
		call	Poll_Push
		movlw	D'13'
		call	Poll_Push
		goto	Sim_Cmd_In
; *********************************************************************
Poll_Push								; char passed in W
		Bank_1
PP_0
		btfss	TXSTA,TRMT
		goto	PP_0					; diddle till ready
		nop
		Bank_0
		movwf	TXREG					; away it goes
		return
; *********************************************************************
Poll_Get
		btfss	PIR1,RCIF				; check for a char
		goto	Poll_Get
		movfw	RCREG					; grab a char
		movwf	TXREG					; echo it
		return
; *********************************************************************
; **********************************************************************
;                       LCD Functions
; **********************************************************************		
;=======================================================================
; LCDINIT
; Initilize LC-Display Module - basic boilerplate setup
; NOTE: Power-up delay not necessary here as the PIC spends time 
;       checking for a bootload prior to running user code!
;       Apply 15mS delay if no BL process!!!!!!
;=======================================================================
LCDINIT
							
		clrf	LCD_CTRL	; ALL PORT output should output Low.
		movlw	0x038		; 8-bit-interface, 2-lines
		call	LCDPUTCMD	; issue cmd
		movlw	0x00		; disp.off, curs.off, no-blink (or enable!)
		call	LCDDMODE
		call	LCDCLEAR
		movlw	0x04		; disp.ON, curs.OFF, curs NON BLINK
		call	LCDDMODE
		movlw	0x02		; auto-inc (shift-cursor)
		call	LCDEMODE
		return				; see ya!
;=============================================================================
; LCD_ENABLE
; Pulses LCD enable pin
;=============================================================================
LCD_ENABLE
		bsf		LCD_CTRL, LCD_E		; LCD E-line High
		bcf		LCD_CTRL, LCD_E		; LCD E-line Low
		return
;=============================================================================
; LCDBUSY
; Returns when LCD busy-flag is inactive
;=============================================================================
LCDBUSY
		bsf			STATUS,RP0			; Select Register page 1
		movlw		0xFF				; Set PORTB for input (just here)
		movwf		LCD_DATA_TRIS
		bcf			STATUS, RP0			; page 0
		bcf			LCD_CTRL, LCD_RS	; Set LCD for command mode
		bsf			LCD_CTRL, LCD_RW	; Setup to read busy flag
		bsf			LCD_CTRL, LCD_E		; LCD E-line High
		movf		LCD_DATA, W			; Read busy flag + DDram address
		bcf			LCD_CTRL, LCD_E		; LCD E-line Low
		andlw		0x80				; Check Busy flag, High = Busy
		btfss		STATUS, Z
		goto		LCDBUSY				; gotta wait it out!
LCDNOTBUSY
		bcf			LCD_CTRL, LCD_RW
		bsf			STATUS, RP0			;  page 1
		movlw		0x00
		movwf		LCD_DATA_TRIS		; Set PORTB back to output
		bcf			STATUS, RP0			; page 0
		return
;=============================================================================
; LCDCLEAR
; Clears display and returns cursor to home position (upper-left corner).
;
;=============================================================================
LCDCLEAR
		movlw		0x01
		call		LCDPUTCMD
		return
;=============================================================================
; LCDHOME
; Returns cursor to home position.
; Returns display to original position (when shifted).
;
;=============================================================================
LCDHOME
		movlw		0x02
		call		LCDPUTCMD
		return
;=============================================================================
; LCDEMODE
; Sets entry mode of display.
; Required entry mode must be set in W
;  b0	: 0 = no display shift	1 = display shift
;  b1	: 0 = auto-decrement	1 = auto-increment
;  b2-7	: don't care
;=============================================================================
LCDEMODE
		andlw	0x03		; Strip upper bits
		iorlw	0x04		; Function set
		call	LCDPUTCMD
		return
;=============================================================================
; LCDDMODE
; Sets display control.
; Required display mode must be set in W
;  b0	: 0 = cursor blink off	1 = cursor blink on
;  b1	: 0 = cursor off	1 = cursor on
;  b2	: 0 = display off	1 = display on (display data remains in DDRAM)
;  b3-7	: don't care
;=============================================================================
LCDDMODE
		andlw	0x07		; Strip upper bits
		iorlw	0x08		; Function set
		call	LCDPUTCMD
		return
;=============================================================================
; LCDSCGA
; Sets Character-Generator-RAM address. CGRAM is read/written after
;  this setting.
; Required CGRAM address must be set in W
;  b0-5	: required CGRAM address
;  b6-7	: don't care
;
;=============================================================================
LCDSCGA
		andlw	0x03F		; Strip upper bits
		iorlw	0x040		; Function set
		call	LCDPUTCMD
		return
;=============================================================================
; LCDSDDA
; Sets the Display-Data-RAM address. DDRAM data is read/written after
;  this setting.
; Required DDRAM address must be set in W
;  b0-6	: required DDRAM address
;  b7	: don't care
;=============================================================================
LCDSDDA
		iorlw	0x080		; Function set
		call	LCDPUTCMD
		return
;=============================================================================
; LCDGADDR
; Returns address counter contents, used for both DDRAM and CGRAM.
; RAM address is returned in W
;
;=============================================================================
LCDGADDR
		bsf		STATUS,RP0			; Page 1
		movlw	0x0FF				; Set PORTB for input
		movwf	LCD_DATA_TRIS
		bcf		STATUS, RP0			; Page 0
		bcf		LCD_CTRL, LCD_RS	; Set LCD for command mode
		bsf		LCD_CTRL, LCD_RW	; Setup to read busy flag
		bsf		LCD_CTRL, LCD_E		; LCD E-line High
		movf	LCD_DATA, W			; Read busy flag + RAM address
		bcf		LCD_CTRL, LCD_E		; LCD E-line Low
		andlw	0x07F				; Strip 
		bcf		LCD_CTRL, LCD_RW
		bsf		STATUS, RP0			; Page 1
		movlw	0x000
		movwf	LCD_DATA_TRIS		; Set PORTB for output
		bcf		STATUS, RP0			; Page 0
		return
;=============================================================================
; LCDPUTCHAR
; Sends character to LCD
; Required character must be in W
;=============================================================================
LCDPUTCHAR
		movwf	LCD_TEMP			; Character -> w
		call	LCDBUSY				; Wait for LCD to be ready
		bcf		LCD_CTRL, LCD_RW	; Set LCD in read mode
		bsf		LCD_CTRL, LCD_RS	; Set LCD in data mode
		bsf		LCD_CTRL, LCD_E		; LCD E-line High
		movf	LCD_TEMP, W
		movwf	LCD_DATA			; Send data to LCD
		bcf		LCD_CTRL, LCD_E		; LCD E-line Low
		return
;=============================================================================
; LCDPUTCMD
; Sends command to LCD (strobe E)
; Required command must be in W
;=============================================================================
LCDPUTCMD
		movwf	LCD_TEMP			; Command -> w
		call	LCDBUSY				; Wait for LCD to be ready
		bcf		LCD_CTRL, LCD_RW	; Set LCD in read mode
		bcf		LCD_CTRL, LCD_RS	; Set LCD in command mode
		bsf		LCD_CTRL, LCD_E		; LCD E-line High
		movf	LCD_TEMP, W
		movwf	LCD_DATA			; Send data to LCD
		bcf		LCD_CTRL, LCD_E		; LCD E-line Low
		return
; 
; *****************************************************************************
;     			 Math Routines and Conversions
; *****************************************************************************
; FOLLOWING CONVERSION: TU Peter Hemsley from the web!!

; 5 digit decimal to 16 (17) bit binary. By Peter Hemsley, March 2003.
; Input decimal digits in D0 (LSD) to D4 (MSD)
; Output 16 bit binary in NUMHI and NUMLO
; No temporary variables required
; Code size: 33 instructions
; Execution time: 33 cycles (excluding Call and Return)
; Returns carry set if > 65535 (and NUMHI-LO MOD 65536)
; VERY NICE says BatWing (jwm)
;
; NOTE: Variables are all in COMMON !!

dec2bin16

    movf  	D1,W        ; (D1 + D3) * 2
    addwf 	D3,W
    movwf 	NUMLO
    rlf   	NUMLO,F

    swapf 	D2,W        ; + D2 * 16 + D2
    addwf 	D2,W
    addwf 	NUMLO,F

    rlf   	D4,W        ; + (D4 * 2 + D3) * 256
    addwf 	D3,W
    movwf 	NUMHI

    rlf   	NUMLO,F     ; * 2
    rlf   	NUMHI,F

    swapf 	D3,W        ; - D3 * 16
    subwf 	NUMLO,F
    skpc
    decf  	NUMHI,F

    swapf 	D2,W        ; + D2 * 16 + D1
    addwf 	D1,W
    addwf 	NUMLO,F
    skpnc
    incf  	NUMHI,F

    swapf 	D4,W        ; + D4 * 16 + D0
    addwf 	D0,W

    rlf   	NUMLO,F     ; * 2
    rlf   	NUMHI,F

    addwf 	NUMLO,F
    skpnc
    incf  	NUMHI,F

    movf  	D4,W        ; - D4 * 256
    subwf	 NUMHI,F

    swapf 	D4,W        ; + D4 * 16 * 256 * 2
    addwf 	NUMHI,F
    addwf 	NUMHI,F

    return           
; ***********************************************************************
;
;					Triple Precision ADD By jwm
; ***********************************************************************
Triple_Add
		Bank_2					; to Bank 2
		bcf		STATUS,C
		bcf		STATUS,DC
		movfw	NUMLO
		addwf	Trip0,f
		addcf	Trip1,f
		movfw	NUMHI
		addwf	Trip1,f
		addcf	Trip2,f
		return
; ************************************************************************
;
;					Sum Moving Avg Acc  By jwm
; ************************************************************************
Add_8
		Bank_2					; Bank 2
		clrf		Trip0		; clean
		clrf		Trip1		; it
		clrf		Trip2		; first
;
		movfw		FSR			; save callers FSR
		movwf		Add_FSR_Sav	; save this
		movfw		MA_Ptr_H	; Head of the table
		movwf		FSR
		movlw		D'08'		; loop ctl
		movwf		M_Temp_0	; for 8 double words
Add_8_L
		movfw		INDF		; fetch a MS byte
		movwf		NUMHI		; to MS acc
		incf		FSR			; point LS source
		movfw		INDF		; get it
		movwf		NUMLO		; stash it
		call		Triple_Add	; build into triple prec result
		incf		FSR
		decfsz		M_Temp_0	; doing 8 pairs
		goto		Add_8_L		; continue.
		movfw		Add_FSR_Sav	; restore
		movwf		FSR			; callers FSR
		return
;
; **************************************************************************
;				Make Moving Average  (divide by 8) By jwm
; *************************************************************************
Mak_Mov_Avg
		Bank_2					; insure Bank 2
		movlw		D'3'		; gonna rotate 3 times
		movwf		M_Temp_0
MMA_L
		clrc					; C = 0
		rrf			Trip0,f		; LS first
		clrc
		rrf			Trip1,f		; now middle
		skpnc					; process C into LS if necessary
		bsf			Trip0,7		; we needed it
		clrc
		rrf			Trip2,f		; now MS
		skpnc
		bsf			Trip1,7		; adjust as necessary
		decfsz		M_Temp_0
		goto		MMA_L		; 3 passes = n/8
		return
;
; *************************************************************************
;				Insert to MA Accumulator  By jwm
; *************************************************************************
MA_Process
		Bank_2					; insure Bank 2
		bsf			STATUS,IRP
;
		movfw		MA_Ptr_W	; get running ptr (slot position)
		movwf		FSR
		movfw		NUMHI		; MS result of converted position
		movwf		INDF
		incf		FSR			; LS
		btfsc		NUMLO,0		; make
		incf		NUMLO,f		; even if necessary
		movfw		NUMLO
		movwf		INDF		; in it goes
		call		Add_8		; produce current MA
		incf		FSR			; MS slot for next call
		movfw		MA_Ptr_E	; Last slot+1
		subwf		FSR,w		; are we there?
		skpnz
		goto		MA_Rewind	; yes, back to Slot 0
		movfw		FSR			; not yet,
MA_X
		movwf		MA_Ptr_W	; update running ptr for next pass
		movwf		MA_Ptr_W_S	; and stash it

		if			NOMA
		nop						; for NO MA requested - see NOMA definition
		else
		call		Mak_Mov_Avg	; create a MA for conversion
		endif

		return
;
MA_Rewind
		movfw		MA_Ptr_H	; rewind to the oldest slot for next
		goto		MA_X		; bye
;
; *********************************************************************
;					Binary to Ascii Converter
; *********************************************************************		
; preconditions
	;   TEM0 contains least significant 8 bits of 16-bit operand
	;   TEM2 contains most significant 8 bits of 16-bit operand
	;   the 16-bit operand is positive (if it was negative, it
	;     already been negated and the sign printed)

	; resource usage
	;   TEM0 holds low nibble, is converted to low decimal digit
	;   TEM1 holds second nibble, is converted to second digit
	;   TEM2 holds third nibble, converted to third digit
	;   TEM3 holds 4th nibble, converted to 4th digit
	;   TEMPH short-term scratch
	;   TEMPL short-term scratch
	;
   
	; 	copyright 2002, Douglas W. Jones
	;   this code may be incorporated into any product so long
	;   this notice is preserved in the source code and so long
	;   as any copyright notice on the final software or firmware
	;   product gives credit to the author.

; NOTES by jwm:  25 uS maximum!
; Pass only the args, all memory cleared as necessary here!
; Converted is BCD, must add 0x30 to each to make ASCII
; Leading results zero for smaller values.  OUTSTANDING, thanks!!
;
  
B2A
        SWAPF   TEM0,W          ; get high nibble of low byte
        MOVWF   TEM1            ;   into position
        SWAPF   TEM2,W          ; get high nibble of high byte
        MOVWF   TEM3            ;   into position
        MOVLW   0x0F            ; setup to mask nibbles
        ANDWF   TEM0,F		;   in each nibble, use only low 4 bits
        ANDWF   TEM1,F
        ANDWF   TEM2,F
        ANDWF   TEM3,F

        ; TEM3 ... TEM0 hold 4 nibbles of number, TEM0 holds LSB
        ; TEM0 = (TEM3 + TEM2 + TEM1)*6 + TEM0
        ; TEM0 <=(   8 +   15 +   15)*6 +   15 = 243
        MOVF    TEM3,W
        ADDWF   TEM2,W
        ADDWF   TEM1,W          ; note, C reset because TEM1+TEM2+TEM3 <= 38
        MOVWF   TEMPL           ; TEMPL = W = (TEM3+TEM2+TEM1)
        RLF     TEMPL,F         ; TEMPL =(TEM3+TEM2+TEM1)*2
        ADDWF   TEMPL,F         ; TEMPL =(TEM3+TEM2+TEM1)*3(C reset,TEMPL<=114)
        RLF     TEMPL,W         ; W     =(TEM3+TEM2+TEM1)*6(C reset, W <= 228)
        ADDWF   TEM0,F          ; TEM0 done; TEM0 <= 243, C is still reset

        ; note TEMPH and TEMPL are no-longer needed
        ; TEMPH = TEM0 div 10 is carry into next digit
        ; TEM0  = TEM0 mod 10 is the least significant digit

        ; approximate TEM0/10 by TEM0*.000110011 (assume C already reset)
        MOVF    TEM0,W
        MOVWF   TEMPH           ; TEMPH = TEM0*1.
        RRF     TEMPH,F         ; TEMPH = TEM0*0.1 (assumes C initially zero)
        ADDWF   TEMPH,F         ; TEMPH = TEM0*1.1 (result high bit in C)
        RRF     TEMPH,F         ; TEMPH = TEM0*0.11 (high bit recovered)
        SWAPF   TEMPH,W         ; W     = TEM0*0.000011
        ANDLW   0x0F            ; (discard high bits)
        ADDWF   TEMPH,F         ; TEMPH = TEM0*0.110011 (result high bit in C)
        RRF     TEMPH,F         ; TEMPH = TEM0*0.0110011 (high bit recovered)
        RRF     TEMPH,F         ; TEMPH = TEM0*0.00110011 (bit of junk)
        RRF     TEMPH,F         ; TEMPH = TEM0*0.000110011 (2 bits of junk)
        MOVLW   0x3F            ; (discard junk bits)
        ANDWF   TEMPH,F

        ; TEM0 = TEM0 - 10*TEMPH, the remainder approximation
        RLF     TEMPH,W         ; high bits of TEMPH are zero, so this clears C
        ANDLW   0xFE            ; (clear any random carry in)
        MOVWF   TEMPL           ; TEMPL = W = TEMPH*2
        RLF     TEMPL,F         ; TEMPL = TEMPH*4
        RLF     TEMPL,F         ; TEMPL = TEMPL*8
        ADDWF   TEMPL,W         ; W = TEMPH*8 + TEMPH*2
        SUBWF   TEM0,F          ; TEM0 = TEM0 - 10*TEMPH

        ; TEMPH may be off by 1 and TEM0 off by 10
        MOVLW   -D'10'
        ADDWF   TEM0,W          ; W = TEM0 - 10 (sets C if TEM0 >= 10)
        BTFSC   STATUS,C
        MOVWF   TEM0            ; if W >= 0, TEM0 = TEM0 - 10
        BTFSC   STATUS,C
        INCF    TEMPH,F         ; if W >= 0, TEMPH++

        ; TEM0 is now the least significant decimal digit!

        ; TEM1 = TEMPH + TEM3*9 + TEM2*5 + TEM1
        ; TEM1 <=   24 +    8*9 +   15*5 +   15 = 186 (carry never set!)
        MOVF    TEMPH,W
        ADDWF   TEM3,W
        ADDWF   TEM2,W
        ADDWF   TEM1,F          ; TEM1 = TEMPH + TEM3*1 + TEM2*1 + TEM1(C reset)
        RLF     TEM3,W
        ADDWF   TEM2,W          ; W    =         TEM3*2 + TEM2*1
        MOVWF   TEMPL
        RLF     TEMPL,F         ; TEMPL =        TEM3*4 + TEM2*2
        RLF     TEMPL,W         ; W    =         TEM3*8 + TEM2*4
        ADDWF   TEM1,F          ; TEM1 = TEMPH + TEM3*9 + TEM2*5 + TEM1
        BTFSC   STATUS,Z
        GOTO    PUTDEC1         ; if rest of number is zero, print just 1 digit

        ; TEMPH = TEM1 div 10 is carry into next digit
        ; TEM1  = TEM1 mod 10 is the next to the least significant digit

        ; approximate TEM1/10 by TEM1*.000110011
        MOVF    TEM1,W
        MOVWF   TEMPH           ; TEMPH = TEM1*1.
        RRF     TEMPH,F         ; TEMPH = TEM1*0.1
        ADDWF   TEMPH,F         ; TEMPH = TEM1*1.1 (high bit is in carry)
        RRF     TEMPH,F         ; TEMPH = TEM1*0.11
        SWAPF   TEMPH,W         ; W     = TEM1*0.000011
        ANDLW   0x0F            ; (discard high bits)
        ADDWF   TEMPH,F         ; TEMPH = TEM1*0.110011
        RRF     TEMPH,F         ; TEMPH = TEM1*0.0110011
        RRF     TEMPH,F         ; TEMPH = TEM1*0.00110011
        RRF     TEMPH,F         ; TEMPH = TEM1*0.000110011
        MOVLW   0x3F            ; (discard high bits)
        ANDWF   TEMPH,F

        ; TEM1 = TEM1 - 10*TEMPH, the remainder approximation
        RLF     TEMPH,W
        ANDLW   0xFE            ; (clear any random carry in)
        MOVWF   TEMPL           ; TEMPL = W = TEMPH*2
        RLF     TEMPL,F         ; TEMPL = TEMPH*4
        RLF     TEMPL,F         ; TEMPL = TEMPL*8
        ADDWF   TEMPL,W         ; W = TEMPH*8 + TEMPH*2
        SUBWF   TEM1,F          ; TEM1 = TEM1 - 10*TEMPH

        ; TEMPH may be off by 1 and TEM1 off by 10
        MOVLW   -D'10'
        ADDWF   TEM1,W          ; W = TEM1 - 10
        BTFSC   STATUS,C
        MOVWF   TEM1            ; if W >= 0, TEM1 = TEM1 - 10
        BTFSC   STATUS,C
        INCF    TEMPH,F         ; if W >= 0, TEMPH++

        ; TEM1 is the next to the least significant decimal digit!

        ; TEM2 = TEMPH + TEM2*2
        ; TEM2 <=   18 +   15*2 = 48 (no carry, and top 2 bits unused too)
        MOVF    TEMPH,W
        ADDWF   TEM2,W
        ADDWF   TEM2,F          ; TEM2 = TEMPH + TEM2*2 (never carry out!)
        MOVF    TEM2,W          ; check if need to print high digits
        IORWF   TEM3,W          ; TEM3 must be rolled into test!
        BTFSC   STATUS,Z
        GOTO    PUTDEC2         ; if rest is zero, print just 2 digits

        ; TEMPH = TEM2 div 10 is carry into high two digits
        ; TEM2  = TEM2 mod 10 is the middle digit

        ; get exact TEM2/10 by TEM2*.0001101 (good because TEM2 < 68)
        MOVF    TEM2,W          ; W     = TEM2*1.0
        MOVWF   TEMPH           ; TEMPH = TEM2*1.0
        RRF     TEMPH,F         ; TEMPH = TEM2*0.1
        RRF     TEMPH,F         ; TEMPH = TEM2*0.01
        BCF     TEMPH,7         ; (discard high bit; next to high already clear)
        ADDWF   TEMPH,F         ; TEMPH = TEM2*1.01
        RRF     TEMPH,F         ; TEMPH = TEM2*0.101
        ADDWF   TEMPH,F         ; TEMPH = TEM2*1.101 (does not set carry!)
        SWAPF   TEMPH,W         ; W     = TEM2*0.0001101
        ANDLW   0x0F            ; (discard high bits)
        MOVWF   TEMPH           ; W = TEMPH = TEM2*0.0001101 = TEM2/10

        ; TEM2 = TEM2 - 10*TEMPH, the remainder
        RLF     TEMPH,W
        MOVWF   TEMPL           ; TEMPL = W = TEMPH*2
        RLF     TEMPL,F         ; TEMPL = TEMPH*4
        RLF     TEMPL,F         ; TEMPL = TEMPL*8
        ADDWF   TEMPL,W         ; W = TEMPH*8 + TEMPH*2
        SUBWF   TEM2,F          ; TEM2 = TEM2 - 10*TEMPH

        ; TEM2 is the middle decimal digit!

        ; TEM3 = TEMPH + TEM3*4
        ; TEM3 <=    8 +    8*4 = 40 (no carry, and top 2 bits unused too)
        BCF     STATUS,C
        RLF     TEM3,F          ; TEM3 =         TEM3*2
        RLF     TEM3,W          ; TEM3 =         TEM3*4
        ADDWF   TEMPH,W         ; W    = TEMPH + TEM3*4 (no carry!)
        BTFSC   STATUS,Z
        GOTO    PUTDEC3         ; if rest is zero, just print bottom 3 digits

        ; TEMPH = W div 10 is the high digit
        ; TEM3  = W mod 10 is the least significant digit

        ; get exact W/10 by W*.0001101 (good because W < 68)
        MOVWF   TEM3            ; TEM3  = W*1.0
        MOVWF   TEMPH           ; TEMPH = TEM3*1.0
        RRF     TEMPH,F         ; TEMPH = TEM3*0.1   (carry in is zero!)
        RRF     TEMPH,F         ; TEMPH = TEM3*0.01
        BCF     TEMPH,7         ; (discard high bit; next to high already clear)
        ADDWF   TEMPH,F         ; TEMPH = TEM3*1.01  (carry out is zero!)
        RRF     TEMPH,F         ; TEMPH = TEM3*0.101
        ADDWF   TEMPH,F         ; TEMPH = TEM3*1.101 (does not set carry!)
        SWAPF   TEMPH,W         ; W     = TEM3*0.0001101
        ANDLW   0x0F            ; (discard high bits)
        MOVWF   TEMPH           ; TEMPH = TEM3*0.0001101

        ; TEM3 = TEM3 - 10*TEMPH, the remainder
        RLF     TEMPH,W         ; W     = TEMPH*2
        MOVWF   TEMPL           ; TEMPL = TEMPH*2
        RLF     TEMPL,F         ; TEMPL = TEMPH*4
        RLF     TEMPL,F         ; TEMPL = TEMPL*8
        ADDWF   TEMPL,W         ; W = TEMPH*8 + TEMPH*2
        SUBWF   TEM3,F          ; TEM3 = TEM3 - 10*TEMPH

        ; TEM3 is the next to most significant decimal digit!
        ; TEMPH is the most significant decimal digit!

        ; conversion to decimal is done! now output digits
; Do the BCD->ASCII conversion here!

PUTDEC1
PUTDEC2
PUTDEC3
		return
; ******************************************************************
		end
