; pwm.asm
; V1.0 for Servo Magazine Navigator project
; Examples of various servo methods follow ISR
; jwm - 2008,2009
;********************************************************************************
		LIST P=16F876A
		INCLUDE <P16F876A.INC>
		__CONFIG _WDT_OFF & _HS_OSC & _BODEN_OFF
		
		ERRORLEVEL -302					; kill off the Banksel messages

;--------------------------------------------------------------------------------
A2D		equ		1					; uncomment for A/D Calibrator

; **************************************************************************
; Variable Memory - use Bank 0 page
; Convert to "CBLOCK" form when complete and happy!!
; **************************************************************************
		radix	dec
		org		0x20					; Start of free variable memory
Page0_B	equ		0x20					; Page 0 Base

Debug		equ		Page0_B						; debug placeholder
Debug1		equ		Page0_B+1					; debug location
;
WSave		equ		Page0_B+2					; ISR
CCSave		equ		Page0_B+3					; context save places
A_Shadow	equ		Page0_B+4
TMR0_PS		equ		Page0_B+5					; Timer 0 "postscaler"
Servo_Pos 	equ		Page0_B+6					; Servo Table slot code
Step_Cnt 	equ		Page0_B+7					; step counter
Step_Cnt_Sav equ	Page0_B+8					; worker
A2D_Steps	equ		Page0_B+9					; # steps till A/D value = 0
A2D_Cycle	equ		Page0_B+10
Select_Test equ		Page0_B+11					; Test selector, B0 drives
AD_Center	equ		Page0_B+12					; discovered servo center position
;
Servo_Tab	equ		Page0_B+16					; Servo run mode values
Servo_Tab_E	equ		Page0_B+23					; Last slot
;
; -------------------------------------------------------------------------------
;
; Servo Position Codes - 1.843 mHz rock
; Produces an offset from center based on PORTB inputs PB 0-2
;
Hard_Left_Abs equ 0x1d					; absolute hard left for calibrator

Hard_Left	equ	0x26					; Position 0 (most left throw)
Inc_1		equ	0x1D					; Position 1 (left ++)		
Inc_2		equ	0x23					; Position 2 (left+)
Inc_3		equ	0x26					; Position 3 ( just left of center)
Center		equ	0x2A					; Position 4 (midships rudder)		
Inc_4		equ	0x2E					; Position 5( just rt. of center)
Inc_5		equ	0x31					; Position 6 (right +)
Inc_6		equ	0x38					; Position 7 (right ++)
Hard_Right	equ	0x2E					; Position 8 (most right throw)
;
Post_Scale	equ	0x04					; ISR postscaler for TMR0
Step_Cnt_Max	equ	0x1C				; max position+1 (count)
GRN_Led		equ	2						; Green LED on RA2 (TMR0)
Yel_Led		equ	5						; LED on RA5 for Cal indicator
A_Step		equ	9						; for A/D value subtractor
PB3			equ	3						; cal cycle complete assert
;________________________________________________________________________________
;
		org		0x0000
		clrf	PCLATH
		goto	Start					; start vector
		org		0x0004					; interrupt vector
		goto	Int
;--------------------------------------------------------------------------------
		org		0x100					; Program Memory
Int		goto	Int_X					; for Bootloader
Start	nop
;
; Initialize I/O and PortC and PWM + TMR0

		bcf		INTCON,GIE				; Ints OFF
;
		BANKSEL	TRISA					; Bank 1
		movlw	0x04					; make PA bit RA0 A, RA2 (LED) D
		movwf	ADCON1					; PA all DIGITAL

		movlw	0x0B					; 0-1,3=Ain, 2,4-5 Digital out
		movwf	TRISA

		movlw	0xFF					; PB 0-7 all INPUTS
		movwf	TRISB					; is all debug digital OUTPUTS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Setup TMR0 and Prescaler, force PB weak pullups

		bcf		OPTION_REG,T0CS				; TMR0 -> timer mode
		bcf		OPTION_REG,PSA				; TMR0 gets the prescaler module
	    bsf		OPTION_REG,PS0				; Set
		bsf		OPTION_REG,PS1				;  %128
		bsf		OPTION_REG,PS2				;    prescaler value.
		bcf		OPTION_REG,7				; Allow PB weak pullups
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                    PWM Setup

		BANKSEL	TRISC					; PORTC
		movlw	0x00					; Bit 2 is PWM output
		movwf	TRISC					; done
;
		BANKSEL	PR2						; Timer 2 for PERIOD count
		movlw	0xFF					; = 256, about 20uS period
		movwf	PR2						; 
;
		BANKSEL	CCP1CON					; ls 2 bits of DUTY CYCLE
		movlw	0x3C					; LS two bits + PWM Enable
		movwf	CCP1CON					; done
		BANKSEL	CCPR1L					; LS Register
		movlw	Center					; ms 8 bits of DUTY CYCLE start position
		movwf	CCPR1L					; done DUTY CYCLE
;
		BANKSEL	T2CON					; set prescaler=16 and TMR2 ON
		movlw	0x06
		movwf	T2CON					; now we're running

		BANKSEL	PORTA					; Primary Bank
		clrf	A_Shadow
		clrf	PORTA					; LEDs OFF Initially
		bsf		INTCON,T0IE				; allow TMR0 int
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialize A/D Converter to RA0 and allow ints (Bank 0)
;
		ifdef	A2D						; ONLY IF DEFINED, else no A/D
		bsf		ADCON0,ADON				; turn on the module
		bcf		PIR1,ADIF				; insure complete bit reset
;		bcf		PIR1,TMR1IF				; also TMR1 reset
		BANKSEL	PIE1
		bsf		PIE1,ADIE				; allow A/D int
;		bsf		PIE1,TMR1IE				; also allow TMR1 int
		BANKSEL	PORTA					; back to Bank 0
		bsf		INTCON,PEIE				; allow peripheral ints
		endif
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
				
; Timer 0 "post scale" countdown for managable TMR0 rate, needed for A/D

		movlw	Post_Scale				; 1.843 mHz clock
		movwf	TMR0_PS					; TRIAL VALUE

		movlw	D'7'					; for servo step table builder
		movwf	Step_Cnt				; initialize servo step control down
		movwf	Step_Cnt_Sav			; center -> up

		clrw
		movwf	A2D_Cycle				; conversion control (in TMR0 ISR)
		movwf	Select_Test				; Servo Test selector
		movwf	Servo_Pos				; Head of Step Table
		bsf		INTCON,GIE				; allow interrupts
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;				SERVO Calibration using A/D
; When powered up, allow the A/D. When satisfied with the rudder servo
; position, ground PB 3 for a moment to lock this position and indicate
; by asserting PA, 5 LED.  Finally, build the servo run table based upon
; the center position.  Enter MAIN with a calibrated table.
; Table looks like this : LLL C RRR
AD_Loop
		movfw	PORTB
		movwf	Select_Test				; to determine if centered
		btfsc	Select_Test,PB3			; happy with center?
		goto	AD_Loop					; no, just wait here - A/D active!
;
; Build The Run Table - Center is slot 3 .
;
		movlw	Servo_Tab				; a(table)
		movwf	FSR
		movlw	D'18'					; Slot 0 = center value -18 (L)
		subwf	AD_Center,w				; produce 1st slot value
AD_0
		movwf	INDF					; build low table, bumping up
		addlw	D'6'
		incf	FSR
		decfsz	Step_Cnt				; and low center
		goto	AD_0
;
; For absolute safety, max slot+1 gets max value so ANY value passed
; from the Navigation Board produces a meaningful value!
;
		decf	FSR
		movfw	INDF
		incf	FSR
		movwf	INDF
;
; Indicate Calibration Complete
;
		bsf		PORTA,Yel_Led			; user confirm LED
		bcf		ADCON0,ADON				; turn off the A/D forever
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialization Complete - now do something !
; Poll PB (0-2) for rudder servo position, enter table with result.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; MAIN LOOP

Loop	
		movlw	HIGH Servo_Tab			; insure latch correct
		movwf	PCLATH	

		movfw	PORTB					; PORTB selects servo position
		andlw	0x07					; only interesting field
		movwf	Select_Test				; save for possible use later
;
; PB bits 0-2 used to enter the rudder position table 				
;
		movlw	Servo_Tab				; a(position table)
		addwf	Select_Test,w			; plus the user offset
		movwf	FSR
		movfw	INDF					; get the PWM value
		movwf	CCPR1L					; establish duty cycle
		goto	Loop					; carry on, forever
		
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;-----------------------------------------------------------------------------------------
; Interrupt Vectors w/common context save
;-----------------------------------------------------------------------------------------
; Determine what interrupted and select appropriate ISR
;
Int_X
		movwf	WSave					; current W saved
		swapf	STATUS,w				; current condition codes
		movwf	CCSave					; saved
 ;
		btfsc	INTCON,T0IF				; our timer0?
		goto	Tmr0ISR					; yes, visit ISR
		btfsc	PIR1,ADIF				; no, A/D interrupt?
		goto	AD_ISR					; yeah, go process it
;		btfsc	PIR1,TMR1IF				; no, TMR1?
;		goto	Tmr1ISR					; yup, process
;
; Unexpected interrupt, die here for debug, Green LED hard on
;
Intdie									; BEING HERE IS VERY BAD THING
		bsf		PORTA,GRN_Led			; flag an issue
		goto	Intdie

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;--------------------------------- TMR0 ISR -----------------------------------------------
; Primarily for A/D control, counts, and LED heartbeat
;------------------------------------------------------------------------------------------
Tmr0ISR
		ifdef	A2D						; ONLY IF DEFINED, else no A/D
;
; Arm the A/D every other cycle for debug
;
		movlw	0x01					; toggle
		xorwf	A2D_Cycle,f				; run
		btfsc	A2D_Cycle,0				; a conversion?
		bsf		ADCON0,GO				; request ONE SHOT conversion
		endif
;
; Got a primary "countdown", toggle PORTA LED status via Shadow register
;
		decf	TMR0_PS					; post scaler
		skpz							; Done, toggle PORTA
		goto	Tmr0_xit				; not, just leave status quo
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;		
		movlw	Post_Scale				; reset 
		movwf	TMR0_PS					; postscaler
		movlw	0x01
		xorwf	A_Shadow,f				; toggle the LED bit
		btfsc	A_Shadow,0				; Turn on?
		goto	Toff					; already on, turn off
		bsf		PORTA,GRN_Led 			; was off, turn ON
		goto	T0						; exit
Toff	bcf		PORTA,GRN_Led 			; was on, turn OFF
T0
; Following routines exercise the PWM in several scenarios.  In sailing
; mode they are not used. Informational only.  So exit the ISR.
		
		goto	Tmr0_xit				; exit ISR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;
; If Select_Test = 1, execute Step_Servo with A/D stepper disarmed.
; else, just exit as the A/D ISR is running.
; Three test routines exist: Five_Step, Incremental Step_Servo, and
; A/D Value.  Pick the desired test. B0 set or cleared in Main Loop.
; NOTE: TURN OFF A/D if not using that test!

		btfsc	Select_Test,0			; B0 = 0, Five Steps
		goto	Step_Servo				; B0 = 1, small increment test
;
; Now Update the Servo Position to the PWM Duty Cycle generator
; Five unique positions, then reset to the beginning.

Five_Step
		movlw	HIGH Table
		movwf	PCLATH	

		movfw	Servo_Pos				; get the current
		btfsc	Servo_Pos,3				; currently = 8 (last slot)?
		goto	Reset_Slot				; yes, reset the slot
		incf	Servo_Pos,f				; No, point next for next pass

;
Tab_Call
		call	Table					; fetch the PWM Value
		movwf	CCPR1L					; new duty cycle
		nop
; Exit the ISR - common path
Tmr0_xit
		bcf		INTCON,T0IF				; dispatch this int
		clrf	TMR0					; clear the counter and ...

ISR_Common_Xit
		swapf	CCSave,w				; pop interrupted status
		movwf	STATUS
		swapf	WSave,f
		swapf	WSave,w
		retfie							; see ya!

Reset_Slot
		clrf	Servo_Pos
		goto	Tab_Call
;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Although this table is not used by the Navigation system, the values
;; are.  The center position (0x2A) has been proven to work with most standard
;; Futaba and Hitek servos.  The A/D calibration routine allows the user to tweak
;; this position, then build the dynamic RAM table with standard offsets from
;; center.
Table
		addwf	PCL,f					; return the PWM value
;
		retlw	Inc_1					; 0 most LEFT throw
		retlw	Inc_2					; 1 ++small
		retlw	Inc_3					; 2 +small
		retlw	Center					; 3 midships rudder
		retlw	Inc_4					; 4 +small
		retlw	Inc_5					; 5 ++small
		retlw	Inc_6					; 6 most RIGHT throw
		retlw	Inc_6					; not used
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Step_Servo
		incf	Step_Cnt				; next position
		movfw	Step_Cnt
		sublw	Step_Cnt_Max			; time to wrap?
		skpnz							; test
		goto	Step_Wrap				; need to wrap and reset PWM
;
		movfw	Step_Cnt
		addlw	Hard_Left
		movwf	CCPR1L					; NEW PWM value	
		
		goto	Tmr0_xit				; see ya

Step_Wrap	
		movlw	0xFF					; start -1
		movwf	Step_Cnt
		movlw	Hard_Left				; initial position
		movwf	CCPR1L					; back to PWM
		goto	Tmr0_xit				; exit
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;						A/D Converter ISR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
AD_ISR
;
		clrf	A2D_Steps				; step counter
		movfw	ADRESL					; get the raw conversion results
;		movwf	PORTB					; latch on Port B pins for debug
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
A2D1
		movlw	A_Step
		subwf	ADRESL					; bump it down
		skpc							; skip if still subtracting
		goto	A2D2					; done counting
		incf	A2D_Steps				; bump running cntr
		goto	A2D1					; ... and continue.
;
; A2D_Steps now reflects how much to bump the initial Duty Cycle
; If Select_Test B0=1, DO NOT mess with the PWM register !!  For Cal Mode,
; allow A/D update.  The calibration pot provides for a range of 1.0 - 2.0 ms.
; Once the desired rudder servo center is found, the center value allows the
; remaining table to be computed for actual running.
;
A2D2
;		btfsc	Select_Test,0			; OK to update?
;		goto	AD_XIT					; NO, just leave
		movfw	A2D_Steps				; initial position
		addlw	Hard_Left_Abs			; + the offset, which is 1.0 ms.
		movwf	CCPR1L					; to the Duty Cycle register
		movwf	AD_Center				; and for table builder
AD_XIT
		bcf		PIR1,ADIF				; dispatch this int
;
		goto	ISR_Common_Xit			; pop context and leave
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

		end