// Nav.c Version 1.0
//This is the NAVIGATION Engine for Mr. Darby
// SERVO Magazine version  jwm 2008-2009
// TU Fred Eddy for the LCD and ISR shell-nice!!  jwm
 
//******************************************************************
#include <p18f2525.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <eep.h>


// Settings for PIC18F2525 per Microchip modified convention

#pragma config OSC = HS, WDT = OFF, PWRT = OFF
#pragma config IESO = OFF, LVP = OFF, BOREN = OFF
#pragma config PBADEN = OFF

//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************

void putch(char c);
char ltoa(long n, char *string);

unsigned char Read_b_eep( unsigned int badd );
void Write_b_eep(unsigned int badd,unsigned char bdata);
void Busy_eep ( void );
 
char recvchar(void);
char ssp_recvchar(void);
char sendchar(char data);
char CharInQueue(void);
char ssp_CharInQueue(void);
void lcd_init(void); 
void lcd_send_byte( char address, char n );
void lcd_send_nibble( char n ); 
void lcd_gotoxy( char x, char y); 
void InterruptHandlerHigh (void);
void init_EUSART(void);
double atof ( const char * s );
float fabs( float x );

double deg_to_hdg(double lat1,double lat2,double lon1, double lon2);
double d_dist( double la1,double la2,double lo1, double lo2);
//int calc_end_pt(void);
double get_PC_input(char *msg, int flag);
int get_PC_string(char *msg);
int but_chk(void);
int trip_in();
int trip_calc(void);
void distance_off_course(void);
void rudder(void);
void EE_Write_Byte(unsigned char cc);
unsigned char EE_Read_Byte();
void base_coordinates(void);
double process_base_coordinates(void);

// Follows is the Segment Table, defining one "leg" of a roundtrip.
// Each array element consists of: Segment heading, segment distance,
// Segment waypoint (ending) L/L pair.
//
// The ending waypoint is used to define the start of the next segment,
// ... then new segment heading and distance define another leg.
// This structure is primed with heading and distance args from the
// PC prior to a voyage.  We then internally compute the round trip
// and verify its viability in getting us home.  Note that the only
// initial L/L required is that of the dock!  Everything else is computed
// knowing the true heading and distance in NM for each leg.

#define SEG_TAB_CNT	4				// elements in segment array

struct segment{
	double s_hdg; double s_dist; double s_e_lat; double s_e_lon;
} segtable[SEG_TAB_CNT];


//******************************************************************
//*	PIC18F2525 RAM Definitions - Globals
//******************************************************************
char bytein,window_bot;
unsigned int tmsecs1,imsecs1;		// for ms timer
unsigned int B_Shadow;				// for Heartbeat and LED activity

double g_F_lat, g_F_lon;			// heading/dist functions (globals)
double g_Real_Time_lat, g_Real_Time_lon; // real time GPS position
double g_Desired_h;					// desired heading (s_hdg from segment structure)
double g_XTD, g_XTD_Temp;			// computed distance off course
double g_Leg_Run_Dist;				// total distance run on a leg
double g_HED_Temp;					// used by waypoint arrival calculator
double ftemp;						// general purpose FP temporary
double pi, radval;
int button_push;					// LCD board button depressions
int g_segment_count;				// # live segments (must be <= SEG_TAB_CNT)
int		g_rudder_displacement;		// maintains rudder displacement (from amidships) magnitude
int     g_servo_sense;				// servo sense for specific model
char	g_ch_in[12];				// input processor buffer
int		g_eeflag;					// EEPROM request flag
unsigned int g_EE_adx;				// EEPROM running address
int		g_rudder_count;				// rudder correction "timer"

//******************************************************************
//*	PIC18F2525 ALIAS DEFINITIONS
//******************************************************************
#define	OFF		0
#define ON		1
//******************************************************************
//*	  LCD DEFINITIONS
//******************************************************************
#define databus	   	LATB			// LCD Bus definitions
#define lcdcntrl   	LATA			// LCD Control+coprocessor control for rudder

// PA bits 5,6,7 : Rudder control signal to coprocessor PWM board
 
#define E 	   		0x04			// PA2 (Pin 4)	LCD Control signals 
#define RW  		0x02			// PA1(Pin 3)  
#define RS			0x01			// PA0 (Pin 2)

// Pushbutton Definitions For the four buttons on the LCD Board (S1 is leftmost button)

#define	S1_Button	0x01			// S1 (leftmost)Respond to prompts
#define	S2_Button	0x02			// S2  Request Sync with GPS Board
#define	S3_Button	0x04			// S3  Start A Trip
#define	S4_Button	0x08			// S4 (rightmost, not used at present)

// ******************************************************************************
// LED Definitions - three LEDs on left side of the LCD module

// Rudder Displacement LED
#define Mid_LED_On	B_Shadow |= 0x04;	\
				databus = 0;			\
				databus  |= B_Shadow
#define Mid_LED_Off	B_Shadow &= 0x0B;	\
				databus = 0;			\
				databus  |= B_Shadow
// Bottom, Yellow, Heartbeat, Waypoint Change
#define	Bot_LED_On	B_Shadow |= 0x01;	\
				databus = 0;			\
				databus |= B_Shadow
#define Bot_LED_Off	&= 0xFE;			\
				databus = 0;			\
				databus |= B_Shadow	
// Top, Red, Sync obtained
#define Top_LED_On	B_Shadow |= 0x02;	\
				databus = 0;			\
				databus |= B_Shadow	
#define Top_LED_Off	B_Shadow &= 0xFD;	\
				databus = 0;			\
				databus |= B_Shadow



#define clrRW		lcdcntrl &= ~RW
#define setRW		lcdcntrl |= RW
#define clrRS		lcdcntrl &= ~RS
#define setRS		lcdcntrl |= RS
#define clrE		lcdcntrl &= ~E
#define setE    	lcdcntrl |= E
				   
#define lcdcls	   lcd_send_byte(0,0x01)
#define	line1	   lcd_gotoxy(1,1)
#define	line2	   lcd_gotoxy(2,1)

char LCD_INIT_STRING[5] = {0x28,0x08,0x01,0x06,0x0E};

#pragma romdata const_table
char lcdmsg_ready[]		= "   NAVIGATION   ";
char TRIP_PROMPT[]      = "S1:PC Trip S2:EE";
char BASE_PROMPT[]		= "S1:PC BASE S2:EE";
char COMP_PROMPT[]      = "S1 To Calc Trip ";
char PC_PROMPT[]		= "Enter Trip Parameters - ? to reenter parameter ";
char LEG_PROMPT[]		= "Seg Cnt? ";
char KBD_MSG[]          = " Now KBD Entry! ";
char DOC_LAT[]          = "Dock LAT?";
char DOC_LON[]          = "Dock LON?";
char DOC_HDG[]          = "Dock HDG?";
char DOC_NM[]           = "Dock NM? ";
char LEG_H[]            = "Leg HDG? ";
char LEG_NM[]           = "Leg NM?  ";
char SERV_PROMPT[]		= "Rev? Y=1 ";

char SYNC_PROMPT[]      = "S2 For GPS Sync!";
char NAV_PROMPT[] 		= "S3 To Navigate! ";


#pragma romdata

///******************************************************************
//*	EUSART/SPI BUFFER DEFINITIONS
//******************************************************************
#define BAUD	19200		//desired baud rate
#define FOSC	20000000	//oscillator frequency

#define DIVIDER (unsigned int) 0x40  // 19.2 with SPBRG High

// Pwrs of 2, 256 maximum
#define USART_RX_BUFFER_SIZE  32     
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )

// Pwrs of 2, 256 max 
#define USART_TX_BUFFER_SIZE  32      
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )
// SSP
#define SSP_RX_BUFFER_SIZE		32
#define SSP_RX_BUFFER_MASK	(SSP_RX_BUFFER_SIZE -1)

char USART_RxBuf[USART_RX_BUFFER_SIZE],USART_TxBuf[USART_TX_BUFFER_SIZE];
char USART_TxHead,USART_TxTail,USART_RxHead,USART_RxTail;

char SSP_RxBuf[SSP_RX_BUFFER_SIZE];
char SSP_RxHead, SSP_RxTail;
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1

#define		TIMER0ON			T0CONbits.TMR0ON = 1
#define		TIMER0OFF			T0CONbits.TMR0ON = 0
#define		enable_TMR0int		INTCONbits.TMR0IE= 1

#define		enable_GLOBALint  	RCONbits.IPEN = 0;				\
							  	INTCONbits.GIE = 1;				\
							  	INTCONbits.PEIE = 1;
#define		disable_GLOBALint 	RCONbits.IPEN = 0;				\
							  	INTCONbits.GIE = 0;				\
							  	INTCONbits.PEIE = 0;
							  	
#define		disable_TMR1int  	PIE1bits.TMR1IE = 0
#define		enable_TMR1int	 	PIE1bits.TMR1IE = 1
#define		TIMER1ON			T1CONbits.TMR1ON = 1
#define		TIMER1OFF			T1CONbits.TMR1ON = 0
			
#define mdelay1(msecdelay)		TIMER1OFF;					\
								PIR1bits.TMR1IF = 0;		\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								TIMER1ON;					\
								while(tmsecs1 < msecdelay);
// *************************************************************************
// Rudder Control -coprocessor commands asserted on PA3 - PA5
// *************************************************************************
#define RB3X	0x08				// PA, bit 3 position
#define RB4X	0x10				// PA, bit 4 position
#define	RB5X	0x20				// PA, bit 5 position

//  Latch coprocessor command bits based on following rudder commands, corresponding to coprocessor table
//			COMMAND			COPROCESSOR	TABLE VALUE
#define	L_aggressive				0
#define L_firm						1
#define L_small						2
#define C_midships					3
#define R_small						4
#define R_firm						5
#define R_aggressive				6
// 

// 000
#define L_aggressive_rudder			LATA &= ~RB5X;			\
									LATA &= ~RB4X;			\
									LATA &= ~RB3X;
// 001
#define L_firm_rudder				LATA &=	~RB5X;			\
									LATA &= ~RB4X;			\
									LATA |=  RB3X;
// 010
#define L_small_rudder				LATA &= ~RB5X;			\
									LATA |=  RB4X;			\
									LATA &=	~RB3X;		
// 011
#define midships_rudder				LATA &= ~RB5X;			\
									LATA |=  RB3X;			\
									LATA |=  RB4X;
// 100
#define R_small_rudder				LATA |=  RB5X;			\
									LATA &=	~RB4X;			\
									LATA &=	~RB3X;

// 101
#define R_firm_rudder				LATA |=  RB5X;			\
									LATA &=	~RB4X;			\
									LATA |=  RB3X;
// 110
#define R_aggressive_rudder			LATA |=  RB5X;			\
									LATA |=  RB4X;			\
									LATA &= ~RB3X;																
									
#define Rudder_Delay	4	// cycles between rudder acuations			
//******************************************************************
//*	putch FUNCTION - Standard "C" stuff
//******************************************************************
void putch(char c)
{
   while(!TXSTAbits.TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;				// load the register
}
// *************************************************************
// Follows forms the HIGH INTERRUPT vector !!
// *************************************************************
// High priority interrupt vector

#pragma code InterruptVectorHigh = 0x08
void
InterruptVectorHigh (void)
{
  _asm
    goto InterruptHandlerHigh		// Hookup the vector forever
  _endasm
}
//******************************************************************
//*	INTERRUPT HANDLER ROUTINE - per Microchip boilerplate for C18 compiler
//******************************************************************

#pragma code
#pragma interrupt InterruptHandlerHigh

void
InterruptHandlerHigh ()

{
  char usart_data,tmphead,tmptail;				// USART
  char ssp_data, ssp_tmphead, ssp_tmptail;		// SPI


	if((PIR1bits.TMR1IF && PIE1bits.TMR1IE))
	{
		PIR1bits.TMR1IF = 0;
		TMR1H = 0xEC;	
		TMR1L = 0x77;
	
		++tmsecs1;
	}

	if((INTCONbits.TMR0IF && INTCONbits.TMR0IE))
	{
		INTCONbits.TMR0IF = 0;			// reset the int
		B_Shadow ^= 0x01;				// toggle the flag, produce the heartbeat
	}	
// SPI
	if((PIR1bits.SSPIF && PIE1bits.SSPIE))
	{
  		 ssp_data = SSPBUF;        	    // read the received SPI data 
                                	    // calculate buffer index 
		  ssp_tmphead = ( SSP_RxHead + 1 ) & SSP_RX_BUFFER_MASK;
		  SSP_RxHead = ssp_tmphead;    // store new index 

	    if ( ssp_tmphead == SSP_RxTail )
		   {
		     // ERROR! Receive buffer overflow 
		   }
                                
		SSP_RxBuf[ssp_tmphead] = ssp_data;   // grab the data

		PIR1bits.SSPIF = 0;				// dispatch int

	}		
// USART

  if(PIR1bits.RCIF)
 {

   usart_data = RCREG;                   // read the received usart_data 
                                	    // calculate buffer index 
   tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK;
  USART_RxHead = tmphead;        		// store new index 

   if ( tmphead == USART_RxTail )
   {
     // ERROR! Receive buffer overflow 
   }
                                
   USART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
 }

  
  if(TXSTAbits.TRMT)
  {
                                // check if all usart_data is transmitted  
   if ( USART_TxHead != USART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
     USART_TxTail = tmptail;      // store new index  

    TXREG = USART_TxBuf[tmptail];  // start transmission
   }
   else
   {
     PIE1bits.TXIE = 0;         // disable TX interrupt  
   }
  }

} // ISR END
//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
// load baud rate divisor for DIVIDER bps with 20MHz clock (jammed for 19,200) 

  SPBRG = DIVIDER;	
  TRISCbits.TRISC7 = 1;		//receive pin
  TRISCbits.TRISC6 = 0;		//transmit pin
  TXSTA = 0x04;				//HIGH speed baud rate !!	
  RCSTA = 0x80;				//enable serial port and serial port pins
  
  USART_RxTail = 0x00;		//flush receive buffer 
  USART_RxHead = 0x00;
  USART_TxTail = 0x00;		//flush transmit buffer
  USART_TxHead = 0x00;
//
  SSP_RxTail = 0x00;		//flush receive buffer 
  SSP_RxHead = 0x00;

//	
  IPR1bits.RCIP = 1;			//receive interrupt = high priority
  IPR1bits.TXIP = 1;			//transmit interrupt = high priority
  PIE1bits.RCIE = 1;			//enable receive interrupt
  PIE1bits.SSPIE= 1;			//enable SSP RX int
  INTCONbits.PEIE = 1;			//enable all unmasked peripheral interrupts
  INTCONbits.GIE = 1;			//enable all unmasked interrupts
  RCSTAbits.CREN = 1;			//enable USART1 receiver
  PIE1bits.TXIE = 0;			//disable USART1 transmit interrupt
  TXSTAbits.TXEN = 1;			//transmitter enabled
}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init(void)
{
 	char init8;

	
  	TRISA = 0b11000000;			// 0-5 PA OUT (LCD CTL+Heartbeat), rudder coprocessor asserts		
   	TRISB = 0b00000000;			// All PB OUT (LCD DATA Bus)		
   	TRISC = 0b10011111;			// 0-4 IN, 5-6 OUT, 7 IN (SPI, EUSART)

// *****************************************************************
// Initialize Rudder Control for midships rudder displacement
// *****************************************************************
	midships_rudder;			// configure PA3-PA5
//******************************************************************
//*	CONFIGURE EEPROM READ/WRITE if we need it!
//******************************************************************
	EECON1 = 0b00000000;		// now used to store round trips
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON0 = 0b00000000;	//turn OFF A2D
	ADCON1 = 0b00001111;	//all digital
//******************************************************************
//*	INITIALIZE THE EUSART
//******************************************************************
	init_EUSART();
//******************************************************************
//*	CONFIGURE AND START TIMER1
//* SET TO OVERFLOW EVERY 1mS
//******************************************************************
	TIMER1OFF;				
	T1CON = 0b00000000;
	TMR1H = 0xEC;
	TMR1L = 0x77;
	TIMER1ON;
// CONFIGURE and start TIMER0
//
// ****************************************************************
// Timer 0 used for heartbeat and slow interval timing

	TIMER0OFF;
	T0CON = 0b10000101;		// 1:64 prescaler to T0
	B_Shadow = 0;			// all LED OFF
	TIMER0ON;

//******************************************************************
//*	ACTIVATE INTERRUPTS
//******************************************************************
	enable_TMR1int;
	enable_TMR0int;
    enable_GLOBALint;
	INTCONbits.GIEH = 1;
//******************************************************************
//******************************************************************
//*	INITIALIZE LCD
//******************************************************************
		lcd_init();
		lcdcls;
		line1;
		for(init8=0;init8<16;init8++)
			lcd_send_byte(1,lcdmsg_ready[init8]);
}
//******************************************************************
//*	GET A CHARACTER FROM RECEIVE BUFFER
//******************************************************************
char recvchar(void)
{
  char tmptail;
                                // wait for incomming usart_data  
  while ( USART_RxHead == USART_RxTail );
                                // calculate buffer index  
  tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK;
  USART_RxTail = tmptail;        // store new index  

  return USART_RxBuf[tmptail];   // return usart_data  
}
//******************************************************************
//*	GET A CHARACTER FROM SPI RECEIVE BUFFER
//******************************************************************
char ssp_recvchar(void)
{
  char ssp_tmptail;
                                // wait for incomming usart_data  
  while ( SSP_RxHead == SSP_RxTail );
                                // calculate buffer index  
  ssp_tmptail = ( SSP_RxTail + 1 ) & SSP_RX_BUFFER_MASK;
  SSP_RxTail = ssp_tmptail;        // store new index  

  return SSP_RxBuf[ssp_tmptail];   // return usart_data  
}

//******************************************************************
//*	SEND A CHARACTER
//******************************************************************
char sendchar(char usart_data)
{
  char tmphead;
                                // calculate buffer index 
  tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == USART_TxTail );
                                // store usart_data in buffer 
  USART_TxBuf[tmphead] = usart_data;
  USART_TxHead = tmphead;        // store new index 

  PIE1bits.TXIE = 1;            // enable TX interrupt 

  return usart_data;
}
//******************************************************************
//*	CHECK FOR CHARACTER IN RECEIVE BUFFER
//******************************************************************
char CharInQueue(void)
{
  return(USART_RxHead != USART_RxTail);
}

//******************************************************************
//*	CHECK FOR CHARACTER IN SPI RECEIVE BUFFER
//******************************************************************
char ssp_CharInQueue(void)
{
  return(SSP_RxHead != SSP_RxTail);
}

//******************************************************************
//*	LCD Functions
//******************************************************************
void lcd_send_nibble( char n ) 
{
	databus &= 0x0F;
	databus |= (n & 0xF0);
	mdelay1(1);
	setE;
	mdelay1(1);
	clrE;
}

void lcd_send_byte( char address, char n ) 
{
      clrE;
      
	  switch (address)
	  {
	  		 case 0:
			 	  clrRS;			// RS sending command
				  break;
			 case 1:
			 	  setRS;			// RS data strobe
				  break;
			 default:
			 	  setRS;			// Just strobe it for data
				  break;	
      }		
	  		mdelay1(1);
		    
      lcd_send_nibble (n);
      lcd_send_nibble (n << 4);
	  		mdelay1(1);

}

void lcd_init(void) 
{
    char j8;
    clrRS;
    clrE;
    clrRW;
	mdelay1(16);
	lcd_send_nibble(0x30);
	mdelay1(6);
	lcd_send_nibble(0x30);
	mdelay1(1);
	lcd_send_nibble(0x30);
	mdelay1(2);
    lcd_send_nibble(0x20);
	for(j8=0;j8<5;++j8)
	{
	mdelay1(2);
       lcd_send_byte(0,LCD_INIT_STRING[j8]);
	}   
}

void lcd_gotoxy( char x, char y) 
{
// where x = lcd row (1,2,3,4) and y = column (1 thru 20)
// This is setup in case we ever get a four row device.

   char address;
   switch (x)
   {
   		  case 1:
		  	   address = 0;
			   break;
		  case 2:
		  	   address = 0x40;
			   break;
		  case 3:
		  	   address = 0x14;
			   break;
		  case 4:
		  	   address = 0x54;
			   break;
		  default:
		  	   address = 0;
   }   
   address += (y-1);
   lcd_send_byte(0,0x80|address);
}

// ********************************************************************
// **************************  EE Functions ***************************
// ********************************************************************
// Update EE with all PC input data to allow a trip to be reconstructed if desired.  Always
// initiate writes at EE adx. 0x0000.  The EE routines always update the EE adx. on exit.
// Caller responsible for initializing adx. via the global g_EE_adx.
//
void EE_Write_Byte(unsigned char cc)		// write a byte
{
	unsigned char ch1;

		Busy_eep();							// insure quiet
		Write_b_eep(g_EE_adx,cc);
		g_EE_adx += 1;						// point next write loc.
}

unsigned char EE_Read_Byte()				// read a byte
{
	unsigned char ch1;

		Busy_eep();							// insure quiet
		ch1 = Read_b_eep(g_EE_adx);
		g_EE_adx += 1;						// point next read loc.
		return(ch1);
}
		
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                  NAVIGATION FUNCTIONS
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

// *****************************************************************
//                  Accept Round Trip Args
// *****************************************************************
// Accept all parameters associated with a trip. Prompt user for details of each
// trip leg.  When inputting from remote PC or terminal, the entire session is also
// written to EEPROM as ASCII, including CR string delimiters.  At the beginning of
// this function, prompt the user (via the LCD) to request that the trip be recreated
// using EEPROM data.  In that case, the actual PC related prompts are still issued,
// but go to space.  The recalled trip is then run as normal.  Presently only the
// most recent PC input trip is available for recall.
//

int trip_in()
{
		int i,leg,f;
		long n;
		double temp;
		unsigned char dummy;
// *************************************************************************
		lcdcls;									// clear the LCD
		line1;

		for(i=0;i<47;i++)
			sendchar(PC_PROMPT[i]);				// initial PC msg
		sendchar(0x0d);
		sendchar(0x0a);

// Now accept all trip segment parameters

		for(i=0;i<16;i++)
			lcd_send_byte(1,KBD_MSG[i]);

		for(i=0;i<9;i++)
			sendchar(LEG_PROMPT[i]);			// Prompt for Leg segments
		sendchar(0x0d);
		sendchar(0x0A);
// ***************************************************************************
	g_EE_adx = 0;								// reset EE adx.
	if(!g_eeflag)					 			// 0: accept PC input for NEW trip
      {
		do{
		if((CharInQueue() != 0))				// wait for Segment response
		  {
			g_ch_in[0] = recvchar();			// Accept arg
			EE_Write_Byte((unsigned char)g_ch_in[0]);	// stash in EE string

			sendchar(g_ch_in[0]);				// Echo char
			g_segment_count = g_ch_in[0]-0x30;  // -> printable
			if(g_segment_count > SEG_TAB_CNT)
				g_segment_count = SEG_TAB_CNT;	// must be < SEG_TAB_CNT (4 legs in trip)
			break;								// continue
		  }
		} while(1);
		if((CharInQueue() == 0));
		   i = recvchar();						// dump the CR associated with above

		EE_Write_Byte((unsigned char)i);		// stash in EE string
	 }
	else										// 1: accept from EE (rerun previous trip)
	 {
		g_ch_in[0] = (char)EE_Read_Byte();	    // get segment count
		g_segment_count = g_ch_in[0]-0x30;		// -> an int
		dummy = EE_Read_Byte();					// purge the CR delimiter
	 }
// ****************************************************************************
 
// Start building the segment structures

		f = 0;									// deg conversion
		segtable[0].s_e_lat = get_PC_input(DOC_LAT,f);	// Dock initial LAT	
		line2;
		lcd_send_byte(1,'L');					// post indicator to LCD

// *************************************************************************
// Now Initial LON as above

		segtable[0].s_e_lon	 = get_PC_input(DOC_LON,f);	// Dock initial LAT	
		lcd_send_byte(1,'l');
// *************************************************************************
// Now Initial HEADING

		segtable[0].s_hdg = get_PC_input(DOC_HDG,f);
		lcd_send_byte(1,'H');
// *************************************************************************
// Finally, Initial DISTANCE in NM

		f = 1;									// request NM-> rad conversion
		segtable[0].s_dist = get_PC_input(DOC_NM,f);
		lcd_send_byte(1,'d');
		f = 0;
// *************************************************************************
// OK, all other segments (legs) require only heading and distance to be input,
// as the L/L point at the waypoint is computed!
//						Input Segment Data
// *************************************************************************
		lcd_send_byte(1,' ');					// delimit line 2 cues
		leg = 1;								// leg cntr
		do{
			sendchar((leg+0x30));				// progress bar
			sendchar(' ');
			segtable[leg].s_hdg = get_PC_input(LEG_H,f);	// get leg heading
			lcd_send_byte(1,(leg+0x30));
//
			sendchar((leg+0x30));
			sendchar(' ');
			f = 1;								// NM-> rad conversion
			segtable[leg].s_dist = get_PC_input(LEG_NM,f); // get leg distance in NM
			lcd_send_byte(1,(leg+0x30));
			leg++;
			f = 0;

		  } while(leg < g_segment_count);

// Prompt for Servo Reversal

		f =3;									// servo code for Input Processor
		g_servo_sense = 0;						// assume normal
		ftemp = get_PC_input(SERV_PROMPT,f);	// returned value n.a. here, a global is set

		return(1);
}
// *****************************************************************
//					PC Input Processor
// *****************************************************************
double get_PC_input(char *msg, int f)
{
	int i;
	double temp;
	int flag = f;							// NM conversion indicator

	
	while(!get_PC_string(msg));				// issue prompt and wait for input, retry if user requested

// If the processing flag = 3, process the servo sense, else process floating point input

		if(flag == 3)
			{
				g_servo_sense = g_ch_in[1]-0x30;
				return(0.0);				// return value meaningless here
			}				 

// Convert to float, then to radians.		// All internal calcs in radians

		temp = atof(g_ch_in);

		if(flag == 0)						// hdg is different radian conversion
			temp*=radval;					// deg->radians
		else
			temp = temp*0.00029089;			// distance in NM -> radians

		return(temp);						// always return radians
}
// *******************************************************************************
// Accept valid input string for get_PC_input function and check for abort request
// *******************************************************************************
// If an abort input char (?) appears in the string, return 0 which will force a 
// reprompt/reaccept operation; else return 1.  The input process can be aborted
// at any point prior to the CR - after CR the string is processed regardless...
// RETURN: a CR delimited input string available in g_ch_in[] .

int get_PC_string(char *msg)
{
		int i;

		g_ch_in[0] = ' ';						
		for(i=0;i<9;i++)
			sendchar(msg[i]);				// Prompt for PC input
		sendchar(0x0d);
		sendchar(0x0A);

// Accept Raw ASCII input from the PC 
		i = 1;								
        if(!g_eeflag)						// g_eeflag 0, accept from PC
        {
		do{
			if((CharInQueue() != 0))				
			  {
				g_ch_in[i] = recvchar();		// Accept arg
				EE_Write_Byte((unsigned char)g_ch_in[i]);	// stash in EE string

				if(g_ch_in[i] == '?')
				  {
					sendchar(0x0d);				// kill off the error
					sendchar(0x0A);				// input string
					return(0);					// will cause this request to be reissued
				  }

				sendchar(g_ch_in[i]);			// Echo char
				i++;
              }          
		  }
		while(g_ch_in[i-1] != 0x0D);			// CR ends input
        }
// Recover previous trip ASCII input to rebuild a trip.
		else									// accept EE data
		{
			do
				{
					g_ch_in[i] = (char)EE_Read_Byte(); // get EE char
					i++;
				}
			while(g_ch_in[i-1] != 0x0D);		// CR ends input
		}

		return (1);								// indicate string OK
}

// *****************************************************************
//						Trip Calculator
// Complete computation for each trip segment in the seg structure
// *****************************************************************
int	trip_calc(void)
{
// Process g_segment_count # of legs, based on the complete "DOCK" starting
// leg.  Calculate the leg endpoint (waypoint), and using this and the
// distance and heading entered for each leg to complete the entire trip.
//
	extern double g_F_lat, g_F_lon;			// returned end points in radians
	double lat_dum,lon_dum,d_dum,tc_dum;
	int i, leg;

	leg = 1;								// start w/first waypoint
	do{
		lat_dum = segtable[leg-1].s_e_lat;
		lon_dum = segtable[leg-1].s_e_lon;
		d_dum   = segtable[leg-1].s_dist;
		tc_dum  = segtable[leg-1].s_hdg;

//
// The calculated end point becomes the waypoint (beginning coordinates of next leg)
// Calculate new endpoint based on previous L/L, heading and distance.

	g_F_lat = asin(sin(lat_dum)*cos(d_dum)+cos(lat_dum)*sin(d_dum)*cos(tc_dum));
	g_F_lon = fmod(lon_dum-asin(sin(tc_dum)*sin(d_dum) / cos(lat_dum))+ pi, 2.0*pi) - pi;

		segtable[leg].s_e_lat = g_F_lat;		// LAT in
		segtable[leg].s_e_lon = g_F_lon;		// LON in

// and so forth for each defined leg!

		leg++;									// repeat for all available legs

	}while(leg <= g_segment_count);				// till done

	return(1);
}
// *****************************************************************
//						Button Checker
// *****************************************************************
// If S1-S4 button depressed, return its value, else return 0.

int	but_chk(void)
{
	extern int button_push;
	int	raw_buttons;

	button_push = 0;
	raw_buttons = PORTC;			// read 'em
									// debounce
	raw_buttons = PORTC;			// OK

	raw_buttons ^= 0xFF;			// correct logical sense
	raw_buttons &= 0x07;			// only 0-3 interesting

	if(raw_buttons != 0)
		button_push = raw_buttons;	// only return actual push!
	return button_push;
}
//                  End Of Function
// *****************************************************************
// *****************************************************************
// Determine True Heading given two sets of L/L waypoints
// All computation in radians!
// *****************************************************************

double deg_to_hdg( double la1,double la2,double lo1, double lo2)
{
	extern double radval,pi;
	extern double pi;
	double lat1, lat2,lon1,lon2;
	double TH,x,y;

	lat1= la1;							// degrees expressed as radians
	lat2= la2;							// broken out here for debug purposes
	lon1= lo1;
	lon2= lo2;
// ***********************************************************************
	y = sin(lon2-lon1)*cos(lat2);
	x = cos(lat1)*sin(lat2)-sin(lat1)*cos(lat2)*cos(lon2-lon1);

// Determine quadrant in radians
// This gets funky right at 90 and 270 degrees (at pi and 2pi), due to the 
// floating precision on this hog leaving a bit of non-zero, but very minute
// residue in the x (LAT) calcs.  Error factor about 0.2 degrees at those edges.

	if(y>0)
		{
			if(x>0)
				TH = atan(y/x);
			else
			if(x<0)
				TH = pi + atan(-y/x);	// 180+ going west of S degrees
			else
				TH = pi/2.0;			// 90 degrees
		}
	else if(y<0)
		{
			if(x>0)
				TH = -atan(-y/x);
			else
			if(x<0)
				TH = atan(y/x)-pi;			
			else
				TH = 3*pi/2;
		}
	else if(y==0)
		{
		if(x>0)
				TH = 2*pi;
		if(x<0)
				TH = pi;
		if(x==0)
				TH = 1;				
		}
// ****************************************************************
// Now adjust for not precise "x" value, by cheating at the edges.

	TH /= radval;						// radian to degree result ( debug!)

	if(TH>1.0 && TH<90.0)
		TH = 360.0-TH;					// adjust for quad 2

	if(TH<1)
		TH*= -1;						// adjust for quads 1, 4

	TH *= radval;						// back to radians (debug)
	return TH;
// ****************************************************************
}//				End of Function
// ****************************************************************
// ****************************************************************
//				DISTANCE BETWEEN L/L POINTS
// Determine distance in NM between two sets of points
// Coordinates passed in radians, distance returned as NM expressed as radians

double d_dist( double la1,double la2,double lo1, double lo2)
{
	extern double radval,pi;
	double lat1, lat2,lon1,lon2;
	double x,y,d,d1,d2,d3;

	lat1= la1;							// degrees expressed as radians
	lat2= la2;							// broken out here for debug purposes
	lon1= lo1;
	lon2= lo2;


	d1 = sin((lat1-lat2)/2.0);
	d1 = d1*d1;

	d2 = sin((lon1-lon2)/2.0);
	d2 = d2*d2;

	d3 = d1+cos(lat1)*cos(lat2)*d2;
	d3 = sqrt(d3);

	d  = 2.0*asin(d3);							// distance in NM (radians)!

	return d;
}					// End Of Function
// ***********************************************************************
// ***********************************************************************
//						Distance Off Course
// ***********************************************************************
// While under weigh, determine how far abeam of the actual desired course we are
// by determining the actual heading and distance travelled to a "snapshot" point
// out in space.  A (+) error distance indicates right of the desired course, a (-)
// distance indicates left of the desired course.
// INPUT:   Coordinates of segment START waypoint (g_F_lat, g_F_lon)
//      :   Coordinates of the real-time GPS snapshot point (g_Real_Time_lat, ..._lon)
//      :   Desired heading from the segment structure (g_Desired_h)
// Returns: g_XTD, distance off course, as the global
// All args expressed in radians
// *********************************************************************** 
// (-) XTD indicates Left of course, (+) indicates right of course
void distance_off_course(void)
{
	double dAD,hAD;						//distance & heading to GPS point

	dAD =d_dist((g_F_lat),(g_Real_Time_lat),(g_F_lon),(g_Real_Time_lon));
	hAD =deg_to_hdg((g_F_lat),(g_Real_Time_lat),(g_F_lon),(g_Real_Time_lon));
	
	g_XTD = asin(sin(dAD)*sin(hAD-(g_Desired_h)));
}
//							End Of Function
// ***********************************************************************
//						Rudder Displacement Processor
// We have determined direction and displacement of the ships rudder, determined by XTD - the distance off
// the desired course.  A (-) value indicates we are LEFT of course, (+) RIGHT of course.  We are
// here because the navagation routines determine an increasing off course heading, which might also
// be a major waypoint arrival correction.  The actual rudder PWM generating engine is an offboard
// coprocessor, previously calibrated to idle at rudder midships position.  The ships rudder in this
// case is large and a small displacement produces emphatic change of course, consequently the co-
// processor displacement table creates small displacements which must be determined by sea trials!
// Three increasing rudder positions are available in each direction.  For a standard course correction,
// we apply the appropriate displacement, held until navigator centers the rudder (back in the slot),
// to stop the turn.  If a radical change of heading is necessary, say at a waypoint, a more vigorous
// displacement is applied.  This routine is probably the MOST likely candidate for refinement going
// forward...like using a compass sensor to stabalize and nail down a turn to a new heading.  Right
// now, we are very loosly coupled to the navigation PID that produces the off course delta.
// When at a major waypoint course correction, indicated by a displacement value > 1, we HOLD the rudder
// for two seconds (TBD!!), then come amidships.  The navigator will then bring us into the new slot.
//
// The Navigation Board communicates with the PWM rudder coprocessor via three digital lines.  The seven
// possible binary values are translated by the coprocessor. PORTA PA5,PA4,PA3 are the control bits, see
// their definitions at the beginning of the program.
//
// Rudder Displacements passed in 'g_rudder_displacement' : 0-small, 1-firm, 2-aggressive  as described
// in the Waypoint processor.  A value of 0 is always used when running a leg and correction is rqd.
//
// The g_servo_sense global was defined during course entry, to accomodate servos that may by their
// nature be reversed.
// NOTE:  ALL rudder commands are multi-statement MACROS and must be bracketed when used in any 
// conditional blocks or bad things may happen !!
 

void rudder(void)
{
			extern int g_rudder_count;
			int code;

		if(g_rudder_count == 0)
			   g_rudder_count = Rudder_Delay;				// allow correction every 'delay' pass

		  if(g_rudder_count == Rudder_Delay)				// execute rudder on 1st pass
		   {
			if (g_servo_sense)
				g_XTD *= -1.0;								// flip servo sense if required

			if(g_XTD > 0.0)									// >0.0, turn left, quadrant independent
				{
				  code = 3-g_rudder_displacement;			// LEFT side of table
				  if(code == 0)
				   {
					L_aggressive_rudder;
				   }
			  	  else if (code == 1)
				   {
					L_firm_rudder;
				   } 
				  else										// code =2
				   {
					L_small_rudder;
				   } 
// Hold displacement for 1 sec., then midships rudder.
				 mdelay1(2048);								// hold time DEBUGGING !!!
				   {
					R_small_rudder;							// kill turn
				   }
				 mdelay1(512);								// hold time DEBUGGING!!
				 g_rudder_count -= 1;						// rudder() entry control

				   {
					midships_rudder;
				   }
				}

			else												// RIGHT side of table
				{
				  code = 3+g_rudder_displacement;
				  if(code == 4)
				   {
				    R_small_rudder;
				   }     
				  else if (code == 5)
				   {       
    			    R_firm_rudder;
				   } 
				  else
                   {
				    R_aggressive_rudder;
                   } 
// Hold displacement for 1 sec., then midships rudder.
				 mdelay1(2048);								// hold time DEBUGGING!!!
				 {
				   L_small_rudder;							// arrest turn
				 }
				 mdelay1(512);								// hold time DEBUGGING!!!
				 g_rudder_count -= 1;						// rudder() entry control

				   {
					midships_rudder;
				   }
				}

// If code != to "2" or "4" ( the standard turn values) we are at a waypoint and will hold the greater-
// than-standard turn for two seconds, then come midships.  The Navigator will finish bringing us into
// the new slot.  Following entered at waypoint turn.

			if((code != 2 ) && (code != 4))
			  {
				mdelay1(2048);									// hold non-standard turn for ~ 2 sec.


// Now center the rudder and exit
			   {
				midships_rudder;									// back midships
      		   }         
			 }
			}
		   else
			{
				g_rudder_count -= 1;							// if pass =2 or =1, just adjust count
			}													// and exit.			


}
//************************************************************************
// base_coordinates
// Accept user's unique base coordinates, maintained in EE.
// If S2 button, user desires EXISTING base coordinates so just exit.
// If you don't have a base in EE, you may end up on the moon or elsewhere!
//************************************************************************
void base_coordinates(void)
{
	int i,j;
	char msg0[]={"Lat as dd"};
	char msg1[]={"FLat : mm"};
	char msg2[]={"Lon as dd"};
	char msg3[]={"FLon : mm"};

	line2;
	for(j=0;j<16;++j)
       lcd_send_byte(1,BASE_PROMPT[j]);		// request S1 or S2 button
											// for user input or EE extraction

	do{
	if(	(i = but_chk()) == S1_Button)			// S1 ?
		{
			g_eeflag = 0;						// PC input
			g_EE_adx = 0x300;					// EE Target
			while(!get_PC_string(msg1));	// issue prompt and wait for input, retry if user requested
			while(!get_PC_string(msg0));
			while(!get_PC_string(msg3));
			while(!get_PC_string(msg2));

			return;
        }
	else if( (i=but_chk()) == S2_Button)
		{
			return;								// base data already in EE, just leave doing nothing!
		}
    }while(1);									// until valid button
}
//************************************************************************
// Process Base Coordinates
// Convert the Base Coordinates in EE (0x300) from ASCII to FP, return a double.
// NOTE: The global EE address is maintained in the physical EE read function, so
// consecutive calls here always return the NEXT available arg.
//************************************************************************
double process_base_coordinates(void)
{
	int i;
	char EE_txt[8];						// EE->fp conversions

	i=0;
	do{
		EE_txt[i++]=EE_Read_Byte();		// read until CR delimiter, and exit...
	}while(EE_txt[i-1] != 0x0d);		// pointing at NEXT arg in EE.

	return(atof(EE_txt));				// return a double
} 
//************************************************************************
//*	MAIN FUNCTION
//************************************************************************

void main(void)  

{	
// ***********************************************************************************
	extern float g_F_lat, g_F_lon;		// returned end points
	extern int button_push = 0;			// depressed buttons
	extern double radval;				// 1 degree->radians
	extern double pi;
//
	int in_flg = 0, q_flg =0;
	int i,j,k,leg,b,toggle;
	long int	l,m;					// for (XTD display)
	int converted;
	char GPS_txt[20];					// Live L/L MA string from GPS Processor
	char LCD_Buffer[]={' ',' ',' ',' ',' ',' ',' ',' '};				// converted value of hits, DEBUG
	char SS_Data;						// response byte from SPI slave -> GPS Processor
	int	S3_active = 0;					// NAV active flag
	int	eeflag;							// EE trip request flag

	double d;
	double THH;
	double dAD,hAD,XTD;					// used by "amount off course" routines
	double dist_off_course_0, dist_off_course_1;
	double XTD_M;
	double f_temp;

// *********************************************************************************
//	Degrees -> radians : decimal deg. * radval, radians -> decimal deg. : radians/radval

	pi = 3.14159265;
	radval = 0.01745329;				// this is = pi/180.0;
	g_EE_adx = 0x0000;					// input string running adx. for EE operations
	g_rudder_count = 0;					// allow rudder corrections when == 0

// *********************************************************************************
  	init();									// Init the entire system
// *********************************************************************************
// Process Base Coordinates to/from EE (user has entered via PC)

	 base_coordinates();				// one-shot base coordinates selection (PC or EE)

// *********************************************************************************
	mdelay1(512);
	line2;									// position for prompts
//
// If S1 Button, enter trip params from PC or terminal, if S2, recreate a trip from EEPROM
//
	for(j=0;j<16;++j)
       lcd_send_byte(1,TRIP_PROMPT[j]);		// request S1 or S2 button

	do{
	if(	(i = but_chk()) == S1_Button)			// S1 ?
		{
			g_eeflag = 0;						// PC input
			if(j = trip_in() == 1);				// yes, get a trip!
				break;						    // got a trip
        }
	else if( (i=but_chk()) == S2_Button)
		{
			g_eeflag = 1;						// EE data recovery
			if(j = trip_in() == 1);				// yes, reconstruct a trip from EE
				break;						    // got a trip
		}
    }while(1);

// ********************************************************************
// We have trip segment parameters in, complete internal waypoint calcs
// ********************************************************************

// Prompt for Button S1 to begin internal processing

	lcdcls;
	line1;

	for(j=0;j<16;++j)
		lcd_send_byte(1,COMP_PROMPT[j]);	// indicate calculating trip

// All internal processing is is radians

	do{
	if(	(i = but_chk()) == S1_Button)		// S1 ?
		{
			if((j = trip_calc()) == 1);		// yes, process complete round trip
				break;						// got a result !
        }      
    }while(1);

// Sync up with GPS Board - Button S2

	line2;
	for(j=0;j<16;++j)
		lcd_send_byte(1,SYNC_PROMPT[j]);	// Request comm with GPS Board

	do{
	if(	(i = but_chk()) == S2_Button)		// S2 - sync ?
		{
			break;
        }      
    }while(1);

// ************************************************************************************
// Sync UP ! and commence actual navigation.  GPS L/L data is passed every three seconds
// from the GPS board, meaning we need to be waiting at "ssp_CharInQueue()" within the
// three sec. window to avoid crashing sync.  There is plenty of time to do all navigation
// in the window, for sure!  (The top red LED on each processor is on hard when the sync
// process is functioning.)  The TOP ROW (line1) of the LCD displays the L/L received
// each GPS period.  
// ************************************************************************************

	lcdcls;									// clear the LCD

// ************************************************************************************	
// Set SPI as SLAVE to match the SPI TX side on the 16F876A side.  Right now, we are not
// using the PORTA enable bit for the SPI, it is always armed and communicating.  In the
// future, SPI may be multiplexed from the GPS board to allow the compass or other SPI
// peripheral.  If and when, SSPSTAT below needs to = 0x40, necessary when the enable bit
// is utilized.  (Already tested, works)  We manually setup the SPI to avoid the heavy
// memory usage of the very general SPI Library functions.  SS_Data is the byte always
// returned by a SPI slave - we use this to facilitate the sync handshake protocol.

	Top_LED_Off;							// Top LED OFF
											// CKE OFF, NO SSbar
	SSPSTAT = 0x00;							// Config SPI as above ~ was 0x40, no CKE
	SSPCON1 = 0x35;
	SS_Data = 0x20;							// initial SPI return value
// **********************************************************************************
  do
   {
	
	line1;									// reinit for next pass
	q_flg = 0;								// RX LAT when 0, RX LON when 1
	i = 0;
//
	do{
		SSPBUF = SS_Data;
		databus = 0;
		databus |= B_Shadow;				// Update LED status
		 while(!(ssp_CharInQueue()))		// wait SSP input
		  {
			databus = 0;
			databus |= B_Shadow;			// Update all LED status
			if(	(b = but_chk()) == S3_Button)		// S3 ?
			{
				S3_active = 1;				// mark active for Nav loop

	        }      

		  }			
		bytein = ssp_recvchar();			// grab  a char

// Live input processing : accept live data until a cr-Q sequence received

		if (in_flg == 1)
			{
			  if(q_flg == 0)				// if 0, processing LAT
               {
				GPS_txt[i++] = bytein;		// stash
				SS_Data = bytein;			// THIS character gets returned NEXT time
				if(bytein == 'Q')			// last returned was the sequence ender (CR)
					q_flg = 1;				// mark LAT sequence complete, start accepting LON
			   }
			  else
               {
                GPS_txt[i++] = bytein;		// doing LON
				SS_Data = bytein;
				if(bytein == 'Q')
					break;					// **** Enter Running Navigation Routines ****
               }
			}				

// Don't process below after sync acheived, we are continuously receiving live data
// Following only happens ONCE, during initial sync up with GPS Board!

	  if(in_flg != 1)
       {
		if(bytein == 'R')
			in_flg = 1;						// Have sync, time to receive live data

		if(bytein == 0x20)
			{
				line1;						// CHECK THIS, don't really need!
				lcdcls;
			}
		if(bytein == 'Q')					// query and response ??
			{
				SS_Data = '+';				// yes, send back on next a (+)
				Top_LED_On;					// Top LED ON, indicates Sync w/GPS Board
			}
		else
			{
				SS_Data = '-';
			} 				
       }

      }
	while(1);

// A valid input string should be in GPS_txt[] - print it out and stop for now
// Actual Navigation Processing allowed, as we have live data from the GPS Board.

	lcd_send_byte(1,'M');
	lcd_send_byte(1,'A');
	lcd_send_byte(1,' ');


	for(i=0;i<6;i++)							// Output L/L MA on Line 1
	 {
		bytein = GPS_txt[i];
		lcd_send_byte(1,bytein);
	 }
	lcd_send_byte(1,' ');						// Done LAT, process LON
	for(i=8;i<14;i++)
     {
        bytein=GPS_txt[i];
		lcd_send_byte(1,bytein);
     }

// **********************************************************************************
// A current averaged GPS L/L is now available in GPS_txt.  Parse it, reconstitute it
// to the full "dd.ddddd" format, convert to float, convert to radians -> g_Real_Time_lat,
// and  g_Real_Time_lon. Process on, all in radians, of course!
// NOTE: Calculations are "broken out" to allow other base "DOCK" coordinates and debug!
// The BASE COORDINATES are organized in EE as follows, in ASCII:
// 0x300 - start of coordinates
// Use this as a guide : Full LAT 4344.9800  Full LON 07316.4100, understanding that the
// ls m.mmmm is delivered in real time from the GPS.  We provide the rest, based upon actual
// location.  Thus, in EE - 40,43,10,73  (each delimited by a CR)
// This feature can be used for bench-testing a trip by tweaking the "base" args such that
// when combined (below) with the real-time GPS values, a "dock" can be located on your bench!
//
// **********************************************************************************
	GPS_txt[6] = ' ';							// insure delimited
	GPS_txt[14]= ' ';							// for conversions.
	g_Real_Time_lat = atof(GPS_txt);			// now -> float (m.mmmm) partial result
	g_Real_Time_lon = atof(GPS_txt+8);			// do both L/L

// Reconstitute and convert, LAT 

	g_EE_adx = 0x300;							// a(EE Base coordinates)

	f_temp = process_base_coordinates();		// grab the value
	g_Real_Time_lat += f_temp;					// now have mm.mmmm
	g_Real_Time_lat /= 60.0;					// expressed as a fraction

	f_temp = process_base_coordinates();		// grab the value
	g_Real_Time_lat += f_temp;					// now have full dd.dddddd

// Reconstitute and convert, LON

	f_temp = process_base_coordinates();		// grab value
	g_Real_Time_lon += f_temp;					// now have mm.mmmm
	g_Real_Time_lon /= 60.0;					// expressed as a fraction

	f_temp = process_base_coordinates();		// grab value
	g_Real_Time_lon += f_temp;					// now have full dd.dddddd of LON
// ********************************************************************************

	g_Real_Time_lat *= radval;					// Finally, both to radians
	g_Real_Time_lon *= radval;					// ready to go.

// ************************************************************************************
//								Navigation
// To run a course, the FIRST WAYPOINT is the L/L point (s_e_lat, s_e_lon) maintained in
// segment[1].  The heading and distance-to-run are maintained in segment[0].  Thus, at the
// waypoint, the NEXT waypoint is that of segment[leg+1] and the heading, distance are in
// segment[leg].  Navigate until leg == g_segment_count.  Note that the waypoint maintained in
// segment[0] is the original L/L entered from the PC and the waypoint in segment[g_segment_count]
// is actually the start of a phantom leg that is never run, since we are home.  Those two
// coordinate sets should be very close (at the dock!), different due to rounding and algorithm.
// The "sanity check" is based upon those points, assuming that the trip is indeed a round one.
// *************************************************************************************
//						The S3 Button Starts The Nav Loop
//
	if(!S3_active)								// thru here till S3
	{
		line2;
		for(i=0;i<16;i++)
			lcd_send_byte(1,NAV_PROMPT[i]);		// issue request to navigate
		leg = 1;								// first waypoint is within segment[1]
		toggle = 0;								// start with distance off course stage 0
		mdelay1(512);							// ~DEBUG
	} 					
// ************************************************************************************
// Enter the actual nav loop once every cycle (GPS Board provides timebase)
//
  if(S3_active)									// (2) S3 active
   {
	if(1)										// process every pass now!!
		{										// (1)
//
// Determine distance off course.  While testing, apply rudder correction if XTD is > 0.002 NM
// which is about 3.7m.  Sign (-), apply right, else apply left rudder.  Rudder action
// TBD after sea trials!
//
			g_F_lat = segtable[leg].s_e_lat;	// waypoint LAT
			g_F_lon = segtable[leg].s_e_lon;	// waypoint LON
			g_Desired_h = segtable[leg-1].s_hdg; // desired heading from previous segment
			distance_off_course();				// produce g_XTD, the distance off course (signed)
			g_XTD_Temp = g_XTD;					// for NM / Display processing
			g_XTD_Temp *= 3437.74677;			// radians -> nm

// If we wander out of the "slot", apply small corrective rudder, which will be maintained
// until we sail back into the on-course slot.

			if(fabs(g_XTD_Temp) > 0.002)		// ~3.7 meters  TBD in trials
			  {
						g_rudder_displacement = 1;		// "standard rate turn", small displacement
						rudder();						// apply appropriate rudder correction
			  }
			else
			  {
			   {
				midships_rudder;				// midships rudder when in the slot
			   }
				g_rudder_count = 0;
			  }
 
// LCD INFO processor
// Update Line 2 with current navigation status, then
// determine if we are approaching or at a waypoint.
// g_XTD is now expressed in NM
//					
	line2;
	for(m=0;m<16;m++)
		lcd_send_byte(1,' ');					// completely purge Line 2
	line2;

	XTD_M = g_XTD_Temp * 1852.0;				// -> meters for later display
	g_XTD_Temp *= 1000.0;						// scale it for display (XTD NM * 1000)

	for(m=0;m<8;m++)
	  {
		LCD_Buffer[m] = ' ';					// purge buffer
	  }

	l = (long)g_XTD_Temp;						// -> ASCII 
	ltoa(l,LCD_Buffer);


	lcd_send_byte(1,'X');						// NM * 1000 off course indicator
	lcd_send_byte(1,' ');

// Display character can be 0-9 and the (-) sign

	for(m=0;m<8;m++)
		{
			if((LCD_Buffer[m]<0x3A && LCD_Buffer[m]>0x2F)|| (LCD_Buffer[m] == 0x2d))  
			   lcd_send_byte(1,LCD_Buffer[m]);
			else
				break;
		}
// Complete the Line 2 display

	lcd_send_byte(1,'N');						// delimit
	lcd_send_byte(1,'M');
	lcd_send_byte(1,' ');

	l = (long)XTD_M;							// fix XTD in meters
	ltoa(l,LCD_Buffer);							// -> ASCII

	for(m=0;m<4;m++)
		{
			if((LCD_Buffer[m]<0x3A && LCD_Buffer[m]>0x2F)|| (LCD_Buffer[m] == 0x2d))  
			   lcd_send_byte(1,LCD_Buffer[m]);
			else
				break;
		}
	lcd_send_byte(1,'M');						

	lcd_send_byte(1,' ');
	lcd_send_byte(1,'L');
	lcd_send_byte(1,(leg+0x30));				    // display current Leg
	lcd_send_byte(1,(g_rudder_count+0x30));			// + rudder displacement phase

// ****************************************************************************************************
//								WAYPOINT PROCESSOR
// ****************************************************************************************************
// Arrival at or about a waypoint is based upon known distance run and our current real GPS position.
// A heading based soultion is problematic at the short distances we employ as a very small instantaneous
// difference between the desired course and our current heading translates into a large, short term angular
// error near the waypoint on a short run.  See discussion below.  Once at or just past the waypoint, several
// possible rudder displacements are possible based on the required heading change to acquire the new WP.

if(1)											// turn on waypoint processor if desired
{

// *****************************************************************************************************
// Determine distance run between leg ENDING waypoint and current GPS fix.  (The "ending" WP is\
// actually the WP defining the next leg to run.

	g_Leg_Run_Dist = d_dist(g_Real_Time_lat,segtable[leg-1].s_e_lat,g_Real_Time_lon,segtable[leg-1].s_e_lon);

// Determine if our run distance exceeds the desired leg distance by 0.008 NM (~15m).  If so, turn.
// NOTE: this is a worst case calculation.  If a leg of 1.0 NM is run with a consistent error of 5 degrees,
//   we would have run 0.004 NM short of our desired 1.0 NM, or ~7.5m.  Adding another 0.004 NM puts us
//   past the waypoint.  We should hold better than 5 degrees to heading, thus reality should indicate
//   that a break point of 0.008 NM is excessive!

// Following proves out very well under simulation! 1st Sea Trial says 0.008 way to much!
// Modify such that break point is 0.003 anywhere around the L/L breakpoint!

//	if(g_Leg_Run_Dist >= (segtable[leg-1].s_dist+(0.004*0.00029089))) // 0.008 NM -> radians
	if(g_Leg_Run_Dist >= segtable[leg-1].s_dist) 
		{
			Mid_LED_On;								// ~DEBUG - indicate here and stop
			mdelay1(2048);							// ~DEBUG - this will go
			Mid_LED_Off	;							// ~ two sec. flash

			leg++;									// new leg now
			if(leg == g_segment_count+1)
			 {
				do{
					{
					R_firm_rudder;					// firm circle forever at the end ( add your end code here)
					}
					Mid_LED_On;						// Indicate done, stop propulsion in the next revision!
				}while(1);							// done the trip, all stop!
			 }
// Determine which way (right or left) to turn to acquire new heading -- g_Desired_h is now the "old"
// heading (leg[-1], the new heading is found in segtable[current leg]. g_XTD is used to pass the sign
// of the desired turn to rudder().  The sign must be corrected to conform to the leg steering concept
// of being R or L of course...here, positive value means turn R!!

			g_XTD =   segtable[leg-1].s_hdg - g_Desired_h;		// sign of g_XTD indications turn direction
			g_XTD *= -1.0;										// correct this to maintain consistency w/rudder function!

// The headings delta determines magnitude of rudder displacement.
// Convert to degrees for this calculation for easier debug and rudder table tweaks!
// The displacement values allow the rudder() function to pass the desired table offset back to the
// rudder coprocessor.

			ftemp = fabs(g_XTD / radval);					// heading difference in absolute degrees
			if(ftemp <= 30.0)
				g_rudder_displacement = 1;					// small rate standard turn
			else if(ftemp <= 60.0)
				g_rudder_displacement = 2;					// firm rate
			else
				g_rudder_displacement = 3;					// aggressive turn
			g_rudder_count = 0;								// clear any displacement-in-progress, and ...
			rudder();										// begin to acquire new heading
			g_rudder_count = 0;								// reset "timer" following waypoint turn
		}

// Not there yet, sail on!

}
		}										// bottom of the live nav loop (1)
	 }											// bottom of S3 active loop    (2)

   }											// Bottom of the Sync Loop
	while (1);									// forever		
// ************************************************************************************
// END of Main
}				